from typing import Optional, List, Union

from deepeval.utils import get_or_create_event_loop, prettify_list
from deepeval.metrics.utils import (
    construct_verbose_logs,
    get_unit_interactions,
    check_conversational_test_case_params,
    initialize_model,
    a_generate_with_schema_and_extract,
    generate_with_schema_and_extract,
)
from deepeval.test_case import ConversationalTestCase, TurnParams
from deepeval.metrics import BaseConversationalMetric
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.metrics.topic_adherence.template import TopicAdherenceTemplate
from deepeval.metrics.topic_adherence.schema import (
    RelevancyVerdict,
    QAPairs,
    QAPair,
)
from deepeval.metrics.api import metric_data_manager


class TopicAdherenceMetric(BaseConversationalMetric):

    _required_test_case_params = [
        TurnParams.ROLE,
        TurnParams.CONTENT,
    ]

    def __init__(
        self,
        relevant_topics: List[str],
        threshold: float = 0.5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
    ):
        self.relevant_topics = relevant_topics
        self.threshold = 1 if strict_mode else threshold
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.async_mode = async_mode
        self.strict_mode = strict_mode
        self.verbose_mode = verbose_mode

    def measure(
        self,
        test_case: ConversationalTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ):
        check_conversational_test_case_params(
            test_case,
            self._required_test_case_params,
            self,
            False,
            self.model,
            test_case.multimodal,
        )
        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                        _in_component=_in_component,
                        _log_metric_to_confident=_log_metric_to_confident,
                    )
                )
            else:
                unit_interactions = get_unit_interactions(test_case.turns)
                interaction_pairs = self._get_qa_pairs(unit_interactions)
                True_Positives = [0, []]
                True_Negatives = [0, []]
                False_Positives = [0, []]
                False_Negatives = [0, []]
                for interaction_pair in interaction_pairs:
                    for qa_pair in interaction_pair.qa_pairs:
                        qa_verdict: RelevancyVerdict = self._get_qa_verdict(
                            qa_pair
                        )
                        if qa_verdict.verdict == "TP":
                            True_Positives[0] += 1
                            True_Positives[1].append(qa_verdict.reason)
                        elif qa_verdict.verdict == "TN":
                            True_Negatives[0] += 1
                            True_Negatives[1].append(qa_verdict.reason)
                        elif qa_verdict.verdict == "FP":
                            False_Positives[0] += 1
                            False_Positives[1].append(qa_verdict.reason)
                        elif qa_verdict.verdict == "FN":
                            False_Negatives[0] += 1
                            False_Negatives[1].append(qa_verdict.reason)

                self.score = self._get_score(
                    True_Positives,
                    True_Negatives,
                    False_Positives,
                    False_Negatives,
                )
                self.success = self.score >= self.threshold
                self.reason = self._generate_reason(
                    True_Positives,
                    True_Negatives,
                    False_Positives,
                    False_Negatives,
                )

                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Interaction Pairs: \n{prettify_list(interaction_pairs)} \n",
                        "Truth Table:",
                        "\nTrue Positives:",
                        f"Count: {True_Positives[0]}, Reasons: {prettify_list(True_Positives[1])} \n",
                        "\nTrue Negatives: ",
                        f"Count: {True_Negatives[0]}, Reasons: {prettify_list(True_Negatives[1])} \n",
                        "\nFalse Positives: ",
                        f"Count: {False_Positives[0]}, Reasons: {prettify_list(False_Positives[1])} \n",
                        "\nFalse Negatives: ",
                        f"Count: {False_Negatives[0]}, Reasons: {prettify_list(False_Negatives[1])} \n",
                        f"Final Score: {self.score}",
                        f"Final Reason: {self.reason}",
                    ],
                )

                if _log_metric_to_confident:
                    metric_data_manager.post_metric_if_enabled(
                        self, test_case=test_case
                    )

                return self.score

    async def a_measure(
        self,
        test_case: ConversationalTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ):
        check_conversational_test_case_params(
            test_case,
            self._required_test_case_params,
            self,
            False,
            self.model,
            test_case.multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None

        with metric_progress_indicator(
            self,
            async_mode=True,
            _show_indicator=_show_indicator,
            _in_component=_in_component,
        ):
            unit_interactions = get_unit_interactions(test_case.turns)
            interaction_pairs = await self._a_get_qa_pairs(unit_interactions)
            True_Positives = [0, []]
            True_Negatives = [0, []]
            False_Positives = [0, []]
            False_Negatives = [0, []]
            for interaction_pair in interaction_pairs:
                for qa_pair in interaction_pair.qa_pairs:
                    qa_verdict: RelevancyVerdict = self._get_qa_verdict(qa_pair)
                    if qa_verdict.verdict == "TP":
                        True_Positives[0] += 1
                        True_Positives[1].append(qa_verdict.reason)
                    elif qa_verdict.verdict == "TN":
                        True_Negatives[0] += 1
                        True_Negatives[1].append(qa_verdict.reason)
                    elif qa_verdict.verdict == "FP":
                        False_Positives[0] += 1
                        False_Positives[1].append(qa_verdict.reason)
                    elif qa_verdict.verdict == "FN":
                        False_Negatives[0] += 1
                        False_Negatives[1].append(qa_verdict.reason)

            self.score = self._get_score(
                True_Positives, True_Negatives, False_Positives, False_Negatives
            )
            self.success = self.score >= self.threshold
            self.reason = await self._a_generate_reason(
                True_Positives, True_Negatives, False_Positives, False_Negatives
            )

            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Interaction Pairs: \n{prettify_list(interaction_pairs)} \n",
                    "Truth Table:",
                    "\nTrue Positives:",
                    f"Count: {True_Positives[0]}, Reasons: {prettify_list(True_Positives[1])} \n",
                    "\nTrue Negatives: ",
                    f"Count: {True_Negatives[0]}, Reasons: {prettify_list(True_Negatives[1])} \n",
                    "\nFalse Positives: ",
                    f"Count: {False_Positives[0]}, Reasons: {prettify_list(False_Positives[1])} \n",
                    "\nFalse Negatives: ",
                    f"Count: {False_Negatives[0]}, Reasons: {prettify_list(False_Negatives[1])} \n",
                    f"Final Score: {self.score}",
                    f"Final Reason: {self.reason}",
                ],
            )

            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )

            return self.score

    def _generate_reason(self, TP, TN, FP, FN):
        total = TP[0] + TN[0] + FP[0] + FN[0]
        if total <= 0:
            return "There were no question-answer pairs to evaluate. Please enable verbose logs to look at the evaluation steps taken"
        prompt = TopicAdherenceTemplate.generate_reason(
            self.success, self.score, self.threshold, TP, TN, FP, FN
        )
        if self.using_native_model:
            res, cost = self.model.generate(prompt)
            self.evaluation_cost += cost
            return res
        else:
            res = self.model.generate(prompt)
            return res

    async def _a_generate_reason(self, TP, TN, FP, FN):
        prompt = TopicAdherenceTemplate.generate_reason(
            self.success, self.score, self.threshold, TP, TN, FP, FN
        )
        if self.using_native_model:
            res, cost = await self.model.a_generate(prompt)
            self.evaluation_cost += cost
            return res
        else:
            res = await self.model.a_generate(prompt)
            return res

    def _get_score(self, TP, TN, FP, FN) -> float:
        true_values = TP[0] + TN[0]
        total = TP[0] + TN[0] + FP[0] + FN[0]
        if total <= 0:
            score = 0
        else:
            score = true_values / total
        return 0 if self.strict_mode and score < self.threshold else score

    def _get_qa_verdict(self, qa_pair: QAPair) -> RelevancyVerdict:
        prompt = TopicAdherenceTemplate.get_qa_pair_verdict(
            self.relevant_topics, qa_pair.question, qa_pair.response
        )
        return generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=RelevancyVerdict,
            extract_schema=lambda s: s,
            extract_json=lambda data: RelevancyVerdict(**data),
        )

    async def _a_get_qa_verdict(self, qa_pair: QAPair) -> RelevancyVerdict:
        prompt = TopicAdherenceTemplate.get_qa_pair_verdict(
            self.relevant_topics, qa_pair.question, qa_pair.response
        )
        return await a_generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=RelevancyVerdict,
            extract_schema=lambda s: s,
            extract_json=lambda data: RelevancyVerdict(**data),
        )

    def _get_qa_pairs(self, unit_interactions: List) -> List[QAPairs]:
        qa_pairs = []
        for unit_interaction in unit_interactions:
            conversation = "Conversation: \n"
            for turn in unit_interaction:
                conversation += f"{turn.role} \n"
                conversation += f"{turn.content} \n\n"
            prompt = TopicAdherenceTemplate.get_qa_pairs(conversation)
            new_pair = None

            new_pair = generate_with_schema_and_extract(
                metric=self,
                prompt=prompt,
                schema_cls=QAPairs,
                extract_schema=lambda s: s,
                extract_json=lambda data: QAPairs(**data),
            )

            if new_pair is not None:
                qa_pairs.append(new_pair)

        return qa_pairs

    async def _a_get_qa_pairs(self, unit_interactions: List) -> List[QAPairs]:
        qa_pairs = []
        for unit_interaction in unit_interactions:
            conversation = "Conversation: \n"
            for turn in unit_interaction:
                conversation += f"{turn.role} \n"
                conversation += f"{turn.content} \n\n"
            prompt = TopicAdherenceTemplate.get_qa_pairs(conversation)
            new_pair = None

            new_pair = await a_generate_with_schema_and_extract(
                metric=self,
                prompt=prompt,
                schema_cls=QAPairs,
                extract_schema=lambda s: s,
                extract_json=lambda data: QAPairs(**data),
            )

            if new_pair is not None:
                qa_pairs.append(new_pair)

        return qa_pairs

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.score >= self.threshold
            except TypeError:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "Topic Adherence"



import functools
from typing import Callable
from crewai.tools import tool as crewai_tool

from deepeval.tracing.context import current_span_context
from deepeval.tracing.types import ToolSpan


def tool(*args, metric=None, metric_collection=None, **kwargs) -> Callable:
    """
    Simple wrapper around crewai.tools.tool that:
      - prints the original function's input and output
      - accepts additional parameters: metric and metric_collection (unused, for compatibility)
      - remains backward compatible with CrewAI's decorator usage patterns
    """
    crewai_kwargs = kwargs

    # Case 1: @tool (function passed directly)
    if len(args) == 1 and callable(args[0]):
        f = args[0]
        tool_name = f.__name__

        @functools.wraps(f)
        def wrapped(*f_args, **f_kwargs):
            current_span = current_span_context.get()
            if current_span and isinstance(current_span, ToolSpan):
                current_span.metric_collection = metric_collection
                current_span.metrics = metric
            result = f(*f_args, **f_kwargs)
            return result

        return crewai_tool(tool_name, **crewai_kwargs)(wrapped)

    # Case 2: @tool("name")
    if len(args) == 1 and isinstance(args[0], str):
        tool_name = args[0]

        def _decorator(f: Callable) -> Callable:
            @functools.wraps(f)
            def wrapped(*f_args, **f_kwargs):
                current_span = current_span_context.get()
                if current_span and isinstance(current_span, ToolSpan):
                    current_span.metric_collection = metric_collection
                    current_span.metrics = metric
                result = f(*f_args, **f_kwargs)
                return result

            return crewai_tool(tool_name, **crewai_kwargs)(wrapped)

        return _decorator

    # Case 3: @tool(result_as_answer=True, ...) — kwargs only
    if len(args) == 0:

        def _decorator(f: Callable) -> Callable:
            tool_name = f.__name__

            @functools.wraps(f)
            def wrapped(*f_args, **f_kwargs):
                current_span = current_span_context.get()
                if current_span and isinstance(current_span, ToolSpan):
                    current_span.metric_collection = metric_collection
                    current_span.metrics = metric
                result = f(*f_args, **f_kwargs)
                return result

            return crewai_tool(tool_name, **crewai_kwargs)(wrapped)

        return _decorator

    raise ValueError("Invalid arguments")



from typing import Dict, Optional, List
from openai import OpenAI, AsyncOpenAI
from pydantic import SecretStr

from deepeval.errors import DeepEvalError
from deepeval.config.settings import get_settings
from deepeval.models.utils import (
    require_secret_api_key,
    normalize_kwargs_and_extract_aliases,
)
from deepeval.models import DeepEvalBaseEmbeddingModel
from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.constants import ProviderSlug as PS


retry_openai = create_retry_decorator(PS.OPENAI)

valid_openai_embedding_models = [
    "text-embedding-3-small",
    "text-embedding-3-large",
    "text-embedding-ada-002",
]

default_openai_embedding_model = "text-embedding-3-small"

_ALIAS_MAP = {
    "api_key": ["openai_api_key"],
}


class OpenAIEmbeddingModel(DeepEvalBaseEmbeddingModel):

    def __init__(
        self,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        generation_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        normalized_kwargs, alias_values = normalize_kwargs_and_extract_aliases(
            "OpenAIEmbeddingModel",
            kwargs,
            _ALIAS_MAP,
        )

        # re-map depricated keywords to re-named positional args
        if api_key is None and "api_key" in alias_values:
            api_key = alias_values["api_key"]

        if api_key is not None:
            # keep it secret, keep it safe from serializings, logging and alike
            self.api_key: Optional[SecretStr] = SecretStr(api_key)
        else:
            self.api_key = get_settings().OPENAI_API_KEY

        model = model if model else default_openai_embedding_model
        if model not in valid_openai_embedding_models:
            raise DeepEvalError(
                f"Invalid model. Available OpenAI Embedding models: {', '.join(valid_openai_embedding_models)}"
            )
        self.kwargs = normalized_kwargs
        self.generation_kwargs = generation_kwargs or {}
        super().__init__(model)

    @retry_openai
    def embed_text(self, text: str) -> List[float]:
        client = self.load_model(async_mode=False)
        response = client.embeddings.create(
            input=text, model=self.name, **self.generation_kwargs
        )
        return response.data[0].embedding

    @retry_openai
    def embed_texts(self, texts: List[str]) -> List[List[float]]:
        client = self.load_model(async_mode=False)
        response = client.embeddings.create(
            input=texts, model=self.name, **self.generation_kwargs
        )
        return [item.embedding for item in response.data]

    @retry_openai
    async def a_embed_text(self, text: str) -> List[float]:
        client = self.load_model(async_mode=True)
        response = await client.embeddings.create(
            input=text, model=self.name, **self.generation_kwargs
        )
        return response.data[0].embedding

    @retry_openai
    async def a_embed_texts(self, texts: List[str]) -> List[List[float]]:
        client = self.load_model(async_mode=True)
        response = await client.embeddings.create(
            input=texts, model=self.name, **self.generation_kwargs
        )
        return [item.embedding for item in response.data]

    ###############################################
    # Model
    ###############################################

    def load_model(self, async_mode: bool = False):
        if not async_mode:
            return self._build_client(OpenAI)
        return self._build_client(AsyncOpenAI)

    def _build_client(self, cls):
        api_key = require_secret_api_key(
            self.api_key,
            provider_label="OpenAI",
            env_var_name="OPENAI_API_KEY",
            param_hint="`api_key` to OpenAIEmbeddingModel(...)",
        )

        client_kwargs = self.kwargs.copy()
        if not sdk_retries_for(PS.OPENAI):
            client_kwargs["max_retries"] = 0

        client_init_kwargs = dict(
            api_key=api_key,
            **client_kwargs,
        )
        try:
            return cls(**client_init_kwargs)
        except TypeError as e:
            # older OpenAI SDKs may not accept max_retries, in that case remove and retry once
            if "max_retries" in str(e):
                client_init_kwargs.pop("max_retries", None)
                return cls(**client_init_kwargs)
            raise

    def get_model_name(self):
        return f"{self.name} (OpenAI)"



from __future__ import annotations
import uuid
from abc import ABC, abstractmethod

from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    TypedDict,
    TYPE_CHECKING,
    Union,
)
from enum import Enum
from pydantic import BaseModel, ConfigDict

from deepeval.prompt.prompt import Prompt

if TYPE_CHECKING:
    from deepeval.dataset.golden import Golden, ConversationalGolden

PromptConfigurationId = str
ModuleId = str
ScoreVector = List[float]  # scores per instance on D_pareto, aligned order
ScoreTable = Dict[PromptConfigurationId, ScoreVector]

# Type alias for model callback function
ModelCallback = Callable[[Prompt, Union["Golden", "ConversationalGolden"]], str]


@dataclass
class PromptConfiguration:
    id: PromptConfigurationId
    parent: Optional[PromptConfigurationId]
    prompts: Dict[ModuleId, Prompt]

    @staticmethod
    def new(
        prompts: Dict[ModuleId, Prompt],
        parent: Optional[PromptConfigurationId] = None,
    ) -> "PromptConfiguration":
        return PromptConfiguration(
            id=str(uuid.uuid4()), parent=parent, prompts=dict(prompts)
        )


class RunnerStatusType(str, Enum):
    """Status events emitted by optimization runners."""

    PROGRESS = "progress"
    TIE = "tie"
    ERROR = "error"


# Type alias for status callback function
RunnerStatusCallback = Callable[..., None]


class Objective(ABC):
    """Strategy for reducing scores per-metric to a single scalar value.

    Implementations receive a mapping from metric name to score
    (for example, {"AnswerRelevancyMetric": 0.82}) and return a
    single float used for comparisons inside the optimizer.
    """

    @abstractmethod
    def scalarize(self, scores_by_metric: Dict[str, float]) -> float:
        raise NotImplementedError


class MeanObjective(Objective):
    """Default scalarizer: unweighted arithmetic mean.

    - If `scores_by_metric` is non-empty, returns the arithmetic
      mean of all metric scores.
    - If `scores_by_metric` is empty, returns 0.0.
    """

    def scalarize(self, scores_by_metric: Dict[str, float]) -> float:
        if not scores_by_metric:
            return 0.0
        return sum(scores_by_metric.values()) / len(scores_by_metric)


class WeightedObjective(Objective):
    """
    Objective that scales each metric's score by a user-provided weight and sums them.

    - `weights_by_metric` keys should match the names of the metrics passed to the
      metric class names passed to the PromptOptimizer.
    - Metrics not present in `weights_by_metric` receive `default_weight`.
      This makes it easy to emphasize a subset of metrics while keeping
      everything else at a baseline weight of 1.0, e.g.:

          WeightedObjective({"AnswerRelevancyMetric": 2.0})

      which treats AnswerRelevancy as 2x as important as the other metrics.
    """

    def __init__(
        self,
        weights_by_metric: Optional[Dict[str, float]] = None,
        default_weight: float = 1.0,
    ):
        self.weights_by_metric: Dict[str, float] = dict(weights_by_metric or {})
        self.default_weight: float = float(default_weight)

    def scalarize(self, scores_by_metric: Dict[str, float]) -> float:
        return sum(
            self.weights_by_metric.get(name, self.default_weight) * score
            for name, score in scores_by_metric.items()
        )


class AcceptedIterationDict(TypedDict):
    parent: PromptConfigurationId
    child: PromptConfigurationId
    module: ModuleId
    before: float
    after: float


class AcceptedIteration(BaseModel):
    parent: str
    child: str
    module: str
    before: float
    after: float


class PromptConfigSnapshot(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    parent: Optional[str]
    prompts: Dict[str, Prompt]


class OptimizationReport(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    optimization_id: str
    best_id: str
    accepted_iterations: List[AcceptedIteration]
    pareto_scores: Dict[str, List[float]]
    parents: Dict[str, Optional[str]]
    prompt_configurations: Dict[str, PromptConfigSnapshot]



from typing import Any, Optional, List, Dict
from uuid import UUID
from time import perf_counter

from deepeval.tracing.context import current_span_context, current_trace_context
from deepeval.test_case.llm_test_case import ToolCall
from deepeval.tracing.types import (
    LlmOutput,
    LlmToolCall,
)
from deepeval.metrics import BaseMetric
from deepeval.tracing.utils import prepare_tool_call_input_parameters

try:
    from langchain_core.callbacks.base import BaseCallbackHandler
    from langchain_core.outputs import LLMResult
    from langchain_core.outputs import ChatGeneration
    from langchain_core.messages import AIMessage

    # contains langchain imports
    from deepeval.integrations.langchain.utils import (
        parse_prompts_to_messages,
        extract_name,
        safe_extract_model_name,
        safe_extract_token_usage,
        enter_current_context,
        exit_current_context,
    )
    from deepeval.integrations.langchain.patch import tool

    langchain_installed = True
except:
    langchain_installed = False


def is_langchain_installed():
    if not langchain_installed:
        raise ImportError(
            "LangChain is not installed. Please install it with `pip install langchain`."
        )


from deepeval.tracing import trace_manager
from deepeval.tracing.types import (
    LlmSpan,
    RetrieverSpan,
    TraceSpanStatus,
    ToolSpan,
)
from deepeval.telemetry import capture_tracing_integration


class CallbackHandler(BaseCallbackHandler):

    def __init__(
        self,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        thread_id: Optional[str] = None,
        user_id: Optional[str] = None,
        metrics: Optional[List[BaseMetric]] = None,
        metric_collection: Optional[str] = None,
    ):
        is_langchain_installed()
        with capture_tracing_integration("langchain.callback.CallbackHandler"):
            trace = trace_manager.start_new_trace()

            self.trace_uuid = trace.uuid

            trace.name = name
            trace.tags = tags
            trace.metadata = metadata
            trace.thread_id = thread_id
            trace.user_id = user_id
            self.metrics = metrics
            self.metric_collection = metric_collection
            current_trace_context.set(trace)
            super().__init__()

    def on_chain_start(
        self,
        serialized: dict[str, Any],
        inputs: dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        if parent_run_id is None:
            uuid_str = str(run_id)
            base_span = enter_current_context(
                uuid_str=uuid_str,
                span_type="custom",
                func_name=extract_name(serialized, **kwargs),
            )
            base_span.input = inputs
            current_trace_context.get().input = inputs
            base_span.metrics = self.metrics
            base_span.metric_collection = self.metric_collection

    def on_chain_end(
        self,
        output: Any,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        uuid_str = str(run_id)
        base_span = trace_manager.get_span_by_uuid(uuid_str)
        if base_span:
            base_span.output = output
            current_trace_context.get().output = output
            exit_current_context(uuid_str=uuid_str)

    def on_llm_start(
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        uuid_str = str(run_id)
        input_messages = parse_prompts_to_messages(prompts, **kwargs)
        model = safe_extract_model_name(metadata, **kwargs)

        llm_span: LlmSpan = enter_current_context(
            uuid_str=uuid_str,
            span_type="llm",
            func_name=extract_name(serialized, **kwargs),
        )

        llm_span.input = input_messages
        llm_span.model = model
        metrics = metadata.pop("metrics", None)
        metric_collection = metadata.pop("metric_collection", None)
        prompt = metadata.pop("prompt", None)
        llm_span.metrics = metrics
        llm_span.metric_collection = metric_collection
        llm_span.prompt = prompt

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,  # un-logged kwargs
    ) -> Any:
        uuid_str = str(run_id)
        llm_span: LlmSpan = trace_manager.get_span_by_uuid(uuid_str)

        output = ""
        total_input_tokens = 0
        total_output_tokens = 0
        model = None

        for generation in response.generations:
            for gen in generation:
                if isinstance(gen, ChatGeneration):
                    if gen.message.response_metadata and isinstance(
                        gen.message.response_metadata, dict
                    ):
                        # extract model name from response_metadata
                        model = gen.message.response_metadata.get("model_name")

                        # extract input and output token
                        input_tokens, output_tokens = safe_extract_token_usage(
                            gen.message.response_metadata
                        )
                        total_input_tokens += input_tokens
                        total_output_tokens += output_tokens

                    if isinstance(gen.message, AIMessage):
                        ai_message = gen.message
                        tool_calls = []
                        for tool_call in ai_message.tool_calls:
                            tool_calls.append(
                                LlmToolCall(
                                    name=tool_call["name"],
                                    args=tool_call["args"],
                                    id=tool_call["id"],
                                )
                            )
                        output = LlmOutput(
                            role="AI",
                            content=ai_message.content,
                            tool_calls=tool_calls,
                        )

        llm_span.model = model if model else llm_span.model
        llm_span.input = llm_span.input
        llm_span.output = output
        llm_span.input_token_count = (
            total_input_tokens if total_input_tokens > 0 else None
        )
        llm_span.output_token_count = (
            total_output_tokens if total_output_tokens > 0 else None
        )

        exit_current_context(uuid_str=uuid_str)

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        uuid_str = str(run_id)
        llm_span: LlmSpan = trace_manager.get_span_by_uuid(uuid_str)
        llm_span.status = TraceSpanStatus.ERRORED
        llm_span.error = str(error)
        exit_current_context(uuid_str=uuid_str)

    def on_llm_new_token(
        self,
        token: str,
        *,
        chunk,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        **kwargs: Any,
    ):
        uuid_str = str(run_id)
        llm_span: LlmSpan = trace_manager.get_span_by_uuid(uuid_str)
        if llm_span.token_intervals is None:
            llm_span.token_intervals = {perf_counter(): token}
        else:
            llm_span.token_intervals[perf_counter()] = token

    def on_tool_start(
        self,
        serialized: dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        inputs: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        uuid_str = str(run_id)

        tool_span = enter_current_context(
            uuid_str=uuid_str,
            span_type="tool",
            func_name=extract_name(
                serialized, **kwargs
            ),  # ignored when setting the input
        )
        tool_span.input = inputs

    def on_tool_end(
        self,
        output: Any,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,  # un-logged kwargs
    ) -> Any:
        uuid_str = str(run_id)
        tool_span: ToolSpan = trace_manager.get_span_by_uuid(uuid_str)
        tool_span.output = output
        exit_current_context(uuid_str=uuid_str)

        # set the tools called in the parent span as well as on the trace level
        tool_call = ToolCall(
            name=tool_span.name,
            description=tool_span.description,
            output=output,
            input_parameters=prepare_tool_call_input_parameters(
                tool_span.input
            ),
        )
        parent_span = current_span_context.get()
        if parent_span:
            if parent_span.tools_called is None:
                parent_span.tools_called = []

            parent_span.tools_called.append(tool_call)

        trace = current_trace_context.get()
        if trace:
            if trace.tools_called is None:
                trace.tools_called = []

            trace.tools_called.append(tool_call)

    def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,  # un-logged kwargs
    ) -> Any:
        uuid_str = str(run_id)
        tool_span: ToolSpan = trace_manager.get_span_by_uuid(uuid_str)
        tool_span.status = TraceSpanStatus.ERRORED
        tool_span.error = str(error)
        exit_current_context(uuid_str=uuid_str)

    def on_retriever_start(
        self,
        serialized: dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
        **kwargs: Any,  # un-logged kwargs
    ) -> Any:
        uuid_str = str(run_id)
        retriever_span = enter_current_context(
            uuid_str=uuid_str,
            span_type="retriever",
            func_name=extract_name(serialized, **kwargs),
            observe_kwargs={
                "embedder": metadata.get("ls_embedding_provider", "unknown"),
            },
        )
        retriever_span.input = query

    def on_retriever_end(
        self,
        output: Any,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,  # un-logged kwargs
    ) -> Any:
        uuid_str = str(run_id)
        retriever_span: RetrieverSpan = trace_manager.get_span_by_uuid(uuid_str)

        # prepare output
        output_list = []
        if isinstance(output, list):
            for item in output:
                output_list.append(str(item))
        else:
            output_list.append(str(output))

        retriever_span.output = output_list
        exit_current_context(uuid_str=uuid_str)

    def on_retriever_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,  # un-logged kwargs
    ) -> Any:
        uuid_str = str(run_id)
        retriever_span: RetrieverSpan = trace_manager.get_span_by_uuid(uuid_str)
        retriever_span.status = TraceSpanStatus.ERRORED
        retriever_span.error = str(error)
        exit_current_context(uuid_str=uuid_str)



class DeepEvalError(Exception):
    """Base class for framework-originated errors.
    If raised and not handled, it will abort the current operation.
    We may also stringify instances of this class and attach them to traces or spans to surface
    non-fatal diagnostics while allowing the run to continue.
    """


class UserAppError(Exception):
    """Represents exceptions thrown by user LLM apps/tools.
    We record these on traces or spans and keep the overall evaluation run alive.
    """


class MissingTestCaseParamsError(DeepEvalError):
    """Required test case fields are missing."""

    pass


class MismatchedTestCaseInputsError(DeepEvalError):
    """Inputs provided to a metric or test case are inconsistent or invalid."""

    pass



from deepeval.benchmarks.base_benchmark import (
    DeepEvalBaseBenchmark,
    DeepEvalBaseBenchmarkResult,
)
from deepeval.utils import make_model_config
from typing import List, Optional, Dict, Any, Tuple
from tqdm import tqdm
import re
import json

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import (
    DeepEvalBaseBenchmark,
    DeepEvalBaseBenchmarkResult,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.schema import StringSchema
from deepeval.telemetry import capture_benchmark_run


class IFEvalResult(DeepEvalBaseBenchmarkResult):
    model_config = make_model_config(arbitrary_types_allowed=True)
    instruction_breakdown: dict[str, Any]
    predictions: "pd.DataFrame"


class IFEvalInstructionVerifier:
    """
    Verifies instruction compliance for IFEval benchmark.

    Implements rule-based verification for various instruction types including
    punctuation constraints, length constraints, format requirements, and content rules.
    """

    @staticmethod
    def verify_punctuation_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify punctuation-related constraints."""
        if instruction_id == "punctuation:no_comma":
            return "," not in response
        elif instruction_id == "punctuation:no_period":
            return "." not in response
        elif instruction_id == "punctuation:no_question_mark":
            return "?" not in response
        elif instruction_id == "punctuation:no_exclamation_mark":
            return "!" not in response
        return True

    @staticmethod
    def verify_length_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify length-related constraints."""
        if instruction_id == "length_constraints:number_words":
            num_words = kwargs.get("num_words")
            relation = kwargs.get("relation", "exactly")

            if num_words is None:
                return True

            word_count = len(response.split())

            if relation == "exactly":
                return word_count == num_words
            elif relation == "at least":
                return word_count >= num_words
            elif relation == "less than":
                return word_count < num_words
            elif relation == "more than":
                return word_count > num_words

        elif instruction_id == "length_constraints:number_characters":
            num_chars = kwargs.get("num_chars")
            relation = kwargs.get("relation", "exactly")

            if num_chars is None:
                return True

            char_count = len(response)

            if relation == "exactly":
                return char_count == num_chars
            elif relation == "at least":
                return char_count >= num_chars
            elif relation == "less than":
                return char_count < num_chars
            elif relation == "more than":
                return char_count > num_chars

        elif instruction_id == "length_constraints:number_sentences":
            num_sentences = kwargs.get("num_sentences")
            relation = kwargs.get("relation", "exactly")

            if num_sentences is None:
                return True

            sentences = re.split(r"[.!?]+", response)
            sentence_count = len([s for s in sentences if s.strip()])

            if relation == "exactly":
                return sentence_count == num_sentences
            elif relation == "at least":
                return sentence_count >= num_sentences
            elif relation == "less than":
                return sentence_count < num_sentences
            elif relation == "more than":
                return sentence_count > num_sentences

        return True

    @staticmethod
    def verify_format_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify format-related constraints."""
        if instruction_id == "detectable_format:json":
            try:
                json.loads(response)
                return True
            except (json.JSONDecodeError, ValueError):
                return False

        elif instruction_id == "detectable_format:list":
            lines = response.strip().split("\n")
            return (
                len(lines) > 1
                or response.strip().startswith("-")
                or response.strip().startswith("*")
                or response.strip().startswith("1.")
                or response.strip().startswith("•")
            )

        elif instruction_id == "detectable_format:number_bullets":
            num_bullets = kwargs.get("num_bullets")
            if num_bullets is None:
                return True

            bullet_patterns = [r"^\s*[-*•]\s+", r"^\s*\d+\.\s+"]
            bullet_count = 0

            for line in response.split("\n"):
                for pattern in bullet_patterns:
                    if re.match(pattern, line):
                        bullet_count += 1
                        break

            return bullet_count >= num_bullets

        elif instruction_id == "detectable_format:number_highlighted_sections":
            num_highlights = kwargs.get("num_highlights")
            if num_highlights is None:
                return True

            highlight_pattern = r"\*[^*]+\*"
            highlights = re.findall(highlight_pattern, response)
            return len(highlights) >= num_highlights

        elif instruction_id == "detectable_format:title":
            title_pattern = r"<<[^>]+>>"
            return bool(re.search(title_pattern, response))

        return True

    @staticmethod
    def verify_case_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify case-related constraints."""
        if instruction_id == "change_case:english_lowercase":
            return response.lower() == response and response.islower()

        elif instruction_id == "change_case:english_uppercase":
            return response.upper() == response and response.isupper()

        elif instruction_id == "change_case:english_titlecase":
            return response.istitle()

        return True

    @staticmethod
    def verify_startend_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify start/end constraints."""
        if instruction_id == "startend:start_with":
            start_text = kwargs.get("start_text")
            if start_text is None:
                return True
            return response.strip().startswith(start_text)

        elif instruction_id == "startend:end_with":
            end_text = kwargs.get("end_text")
            if end_text is None:
                return True
            return response.strip().endswith(end_text)

        return True

    @staticmethod
    def verify_keywords_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify keyword constraints."""
        if instruction_id == "keywords:must_include":
            required_keywords = kwargs.get("keywords", [])
            response_lower = response.lower()
            return all(
                keyword.lower() in response_lower
                for keyword in required_keywords
            )

        elif instruction_id == "keywords:must_not_include":
            forbidden_keywords = kwargs.get("keywords", [])
            response_lower = response.lower()
            return not any(
                keyword.lower() in response_lower
                for keyword in forbidden_keywords
            )

        return True

    @staticmethod
    def verify_content_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify content-related constraints."""
        if instruction_id == "detectable_content:keyword_frequency":
            keyword = kwargs.get("keyword")
            frequency = kwargs.get("frequency")
            relation = kwargs.get("relation", "exactly")

            if keyword is None or frequency is None:
                return True

            keyword_count = response.lower().count(keyword.lower())

            if relation == "exactly":
                return keyword_count == frequency
            elif relation == "at least":
                return keyword_count >= frequency
            elif relation == "less than":
                return keyword_count < frequency
            elif relation == "more than":
                return keyword_count > frequency

        elif instruction_id == "detectable_content:forbidden_words":
            forbidden_words = kwargs.get("forbidden_words", [])
            for word in forbidden_words:
                if word.lower() in response.lower():
                    return False
            return True

        elif instruction_id == "detectable_content:number_placeholders":
            num_placeholders = kwargs.get("num_placeholders")
            if num_placeholders is None:
                return True

            placeholder_pattern = r"\[[^\]]+\]"
            placeholders = re.findall(placeholder_pattern, response)
            return len(placeholders) >= num_placeholders

        elif instruction_id == "detectable_content:postscript":
            postscript_marker = kwargs.get("postscript_marker", "P.S.")
            return postscript_marker in response

        elif instruction_id == "detectable_content:first_word":
            first_word = kwargs.get("first_word")
            if first_word is None:
                return True

            response_words = response.strip().split()
            return (
                response_words
                and response_words[0].lower() == first_word.lower()
            )

        return True

    @staticmethod
    def verify_structural_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify structural constraints."""
        if instruction_id == "structural_constraints:number_paragraphs":
            num_paragraphs = kwargs.get("num_paragraphs")
            relation = kwargs.get("relation", "exactly")

            if num_paragraphs is None:
                return True

            paragraphs = [p for p in response.split("\n\n") if p.strip()]
            paragraph_count = len(paragraphs)

            if relation == "exactly":
                return paragraph_count == num_paragraphs
            elif relation == "at least":
                return paragraph_count >= num_paragraphs
            elif relation == "less than":
                return paragraph_count < num_paragraphs
            elif relation == "more than":
                return paragraph_count > num_paragraphs

        elif instruction_id == "structural_constraints:number_sections":
            num_sections = kwargs.get("num_sections")
            section_spliter = kwargs.get("section_spliter", "---")

            if num_sections is None:
                return True

            sections = [s for s in response.split(section_spliter) if s.strip()]
            section_count = len(sections)
            return section_count >= num_sections

        return True

    @staticmethod
    def verify_combination_constraints(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> bool:
        """Verify combination constraints."""
        if instruction_id == "combination:repeat_prompt":
            prompt_to_repeat = kwargs.get("prompt_to_repeat")
            if prompt_to_repeat is None:
                return True

            return prompt_to_repeat in response

        return True

    @staticmethod
    def verify_instruction_compliance(
        response: str, instruction_id: str, kwargs: Dict[str, Any]
    ) -> Tuple[bool, str]:
        """Verify compliance with a single instruction."""
        try:
            if instruction_id.startswith("punctuation:"):
                result = (
                    IFEvalInstructionVerifier.verify_punctuation_constraints(
                        response, instruction_id, kwargs
                    )
                )
            elif instruction_id.startswith("length_constraints:"):
                result = IFEvalInstructionVerifier.verify_length_constraints(
                    response, instruction_id, kwargs
                )
            elif instruction_id.startswith("detectable_format:"):
                result = IFEvalInstructionVerifier.verify_format_constraints(
                    response, instruction_id, kwargs
                )
            elif instruction_id.startswith("detectable_content:"):
                result = IFEvalInstructionVerifier.verify_content_constraints(
                    response, instruction_id, kwargs
                )
            elif instruction_id.startswith("structural_constraints:"):
                result = (
                    IFEvalInstructionVerifier.verify_structural_constraints(
                        response, instruction_id, kwargs
                    )
                )
            elif instruction_id.startswith("combination:"):
                result = (
                    IFEvalInstructionVerifier.verify_combination_constraints(
                        response, instruction_id, kwargs
                    )
                )
            elif instruction_id.startswith("change_case:"):
                result = IFEvalInstructionVerifier.verify_case_constraints(
                    response, instruction_id, kwargs
                )
            elif instruction_id.startswith("startend:"):
                result = IFEvalInstructionVerifier.verify_startend_constraints(
                    response, instruction_id, kwargs
                )
            elif instruction_id.startswith("keywords:"):
                result = IFEvalInstructionVerifier.verify_keywords_constraints(
                    response, instruction_id, kwargs
                )
            else:
                return False, f"Unknown instruction type: {instruction_id}"

            reason = f"Instruction '{instruction_id}' {'PASSED' if result else 'FAILED'}"
            return result, reason

        except Exception as e:
            return (
                False,
                f"Error verifying instruction '{instruction_id}': {str(e)}",
            )


class IFEval(DeepEvalBaseBenchmark):
    """
    IFEval (Instruction Following Evaluation) benchmark implementation.

    IFEval is a benchmark for evaluating instruction-following capabilities of language models.
    It tests various aspects of instruction following including format compliance, constraint
    adherence, output structure requirements, and specific instruction types.

    Based on the original IFEval paper: https://arxiv.org/abs/2311.07911
    and implementation: https://github.com/google-research/google-research/tree/master/instruction_following_eval
    """

    def __init__(
        self,
        n_problems: Optional[int] = None,
        verbose_mode: bool = False,
        **kwargs,
    ):
        from deepeval.scorer import Scorer
        import pandas as pd

        super().__init__(**kwargs)
        self.scorer = Scorer()
        self.n_problems = n_problems
        self.verbose_mode = verbose_mode
        self.predictions: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None
        self.instruction_breakdown = None

    def evaluate(self, model: DeepEvalBaseLLM, *args, **kwargs) -> IFEvalResult:
        import pandas as pd

        with capture_benchmark_run("IFEval", self.n_problems or "all"):
            overall_correct_predictions = 0
            overall_total_predictions = 0
            predictions_row = []
            instruction_results = {}

            goldens = self.load_benchmark_dataset()
            if self.n_problems and self.n_problems < len(goldens):
                goldens = goldens[: self.n_problems]

            overall_total_predictions = len(goldens)

            for idx, golden in enumerate(
                tqdm(goldens, desc=f"Processing {len(goldens)} IFEval problems")
            ):
                prediction, score, instruction_scores = self.predict(
                    model, golden
                )
                if score:
                    overall_correct_predictions += 1

                predictions_row.append((golden.input, prediction, score))

                for (
                    instruction_id,
                    instruction_score,
                ) in instruction_scores.items():
                    if instruction_id not in instruction_results:
                        instruction_results[instruction_id] = {
                            "correct": 0,
                            "total": 0,
                        }
                    instruction_results[instruction_id]["total"] += 1
                    if instruction_score:
                        instruction_results[instruction_id]["correct"] += 1

                if self.verbose_mode:
                    self.print_verbose_logs(
                        idx, golden.input, prediction, score, instruction_scores
                    )

            overall_accuracy = (
                overall_correct_predictions / overall_total_predictions
            )
            print(f"Overall IFEval Accuracy: {overall_accuracy:.4f}")

            instruction_accuracies = {}
            for instruction_id, results in instruction_results.items():
                accuracy = results["correct"] / results["total"]
                instruction_accuracies[instruction_id] = accuracy
                print(
                    f"Instruction '{instruction_id}' Accuracy: {accuracy:.4f}"
                )
            predictions: pd.DataFrame = pd.DataFrame(
                predictions_row,
                columns=[
                    "Input",
                    "Prediction",
                    "All_Instructions_Correct",
                ],
            )
            self.predictions = predictions
            self.overall_score = overall_accuracy
            self.instruction_breakdown = instruction_accuracies

            return IFEvalResult(
                overall_accuracy=overall_accuracy,
                instruction_breakdown=instruction_accuracies,
                predictions=predictions,
            )

    def predict(
        self, model: DeepEvalBaseLLM, golden: Golden
    ) -> Tuple[str, bool, Dict[str, bool]]:
        """
        Generate prediction for a single IFEval test case and verify instruction compliance.

        Args:
            model: The language model to evaluate
            golden: The golden test case

        Returns:
            Tuple of (prediction, overall_score, instruction_scores)
        """
        try:
            res: StringSchema = model.generate(
                prompt=golden.input, schema=StringSchema
            )
            prediction = res.answer
        except (TypeError, AttributeError):
            res = model.generate(golden.input)
            prediction = str(res)

        instruction_scores = {}
        all_instructions_passed = True

        metadata = golden.additional_metadata or {}
        instruction_ids = metadata.get("instruction_ids", [])
        kwargs_list = metadata.get("kwargs_list", [])

        for i, instruction_id in enumerate(instruction_ids):
            kwargs = kwargs_list[i] if i < len(kwargs_list) else {}
            passed, reason = (
                IFEvalInstructionVerifier.verify_instruction_compliance(
                    prediction, instruction_id, kwargs
                )
            )
            instruction_scores[instruction_id] = passed
            if not passed:
                all_instructions_passed = False

        return prediction, all_instructions_passed, instruction_scores

    def load_benchmark_dataset(self) -> List[Golden]:
        """
        Load IFEval dataset.

        Returns:
            List of Golden test cases
        """
        from datasets import load_dataset

        if self.dataset:
            dataset = self.dataset
        else:
            dataset = load_dataset("google/IFEval")
            self.dataset = dataset

        goldens: List[Golden] = []

        train_data = dataset["train"]

        for data in train_data:
            prompt = data.get("prompt", "")
            instruction_id_list = data.get("instruction_id_list", [])
            kwargs = data.get("kwargs", [])

            golden = Golden(input=prompt, expected_output="")
            golden.additional_metadata = {
                "instruction_ids": instruction_id_list,
                "kwargs_list": kwargs,
            }

            goldens.append(golden)

        return goldens

    def print_verbose_logs(
        self,
        idx: int,
        input: str,
        prediction: str,
        score: bool,
        instruction_scores: Dict[str, bool],
    ) -> str:
        """
        Print verbose logs for debugging and analysis.

        Args:
            idx: Problem index
            input: Input instruction
            prediction: Model prediction
            score: Overall score (True if all instructions passed)
            instruction_scores: Individual instruction scores

        Returns:
            Formatted verbose log string
        """
        steps = [
            f"Input:\n{input}",
            f"Overall Score: {score}\nPrediction: {prediction[:200]}{'...' if len(prediction) > 200 else ''}",
            "Instruction Breakdown:\n"
            + "\n".join(
                [
                    f"  {inst}: {'✓' if passed else '✗'}"
                    for inst, passed in instruction_scores.items()
                ]
            ),
        ]
        verbose_logs = ""
        for i in range(len(steps) - 1):
            verbose_logs += steps[i]

            if i < len(steps) - 2:
                verbose_logs += " \n \n"

        if self.verbose_mode:
            print("*" * 50)
            print(f"Problem {idx + 1}")
            print("*" * 50)
            print("")
            print(verbose_logs + f"\n \n{steps[-1]}")
            print("")
            print("=" * 70)

        return verbose_logs



# from __future__ import annotations

# from dataclasses import replace
# from typing import List, Any, Union, Optional

# try:
#     from agents import (
#         RunConfig,
#         RunResult,
#         RunResultStreaming,
#         Runner as AgentsRunner,
#     )
#     from agents.agent import Agent
#     from agents.models.interface import ModelProvider
#     from agents.items import TResponseInputItem
#     from agents.lifecycle import RunHooks
#     from agents.memory import Session
#     from agents.run import DEFAULT_MAX_TURNS
#     from agents.run import AgentRunner
#     from agents.run_context import TContext
#     from agents.models.interface import Model
#     from agents.run import SingleStepResult

#     agents_available = True
# except:
#     agents_available = False


# def is_agents_available():
#     if not agents_available:
#         raise ImportError(
#             "agents is required for this integration. Install it via your package manager"
#         )


# from deepeval.tracing.tracing import Observer
# from deepeval.tracing.context import current_span_context, current_trace_context
# from deepeval.tracing.utils import make_json_serializable
# from deepeval.tracing.types import AgentSpan

# # Import observed provider/model helpers from our agent module
# from deepeval.metrics import BaseMetric
# from deepeval.openai_agents.agent import _ObservedModel

# _PATCHED_DEFAULT_GET_MODEL = False
# _PATCHED_DEFAULT_RUN_SINGLE_TURN = False

# def patch_default_agent_runner_get_model():
#     global _PATCHED_DEFAULT_GET_MODEL
#     if _PATCHED_DEFAULT_GET_MODEL:
#         return

#     original_get_model_cm = AgentRunner._get_model
#     try:
#         original_get_model = original_get_model_cm.__func__
#     except AttributeError:
#         original_get_model = original_get_model_cm  # fallback (non-classmethod edge case)

#     def patched_get_model(cls, *args, **kwargs) -> Model:
#         model = original_get_model(cls, *args, **kwargs)

#         agent = kwargs.get("agent") if "agent" in kwargs else (args[0] if args else None)
#         if agent is None:
#             return model

#         if isinstance(model, _ObservedModel):
#             return model

#         llm_metrics = getattr(agent, "llm_metrics", None)
#         llm_metric_collection = getattr(agent, "llm_metric_collection", None)
#         confident_prompt = getattr(agent, "confident_prompt", None)
#         return _ObservedModel(
#             inner=model,
#             llm_metric_collection=llm_metric_collection,
#             llm_metrics=llm_metrics,
#             confident_prompt=confident_prompt,
#         )

#     # Preserve basic metadata and mark as patched
#     patched_get_model.__name__ = original_get_model.__name__
#     patched_get_model.__doc__ = original_get_model.__doc__

#     AgentRunner._get_model = classmethod(patched_get_model)
#     _PATCHED_DEFAULT_GET_MODEL = True


# # if agents_available:
#     # patch_default_agent_run_single_turn()
#     # patch_single_turn_streamed()
#     # patch_default_agent_runner_get_model()


# class Runner(AgentsRunner):

#     @classmethod
#     async def run(
#         cls,
#         starting_agent: Agent[TContext],
#         input: Union[str, list[TResponseInputItem]],
#         *,
#         context: Optional[TContext] = None,
#         max_turns: int = DEFAULT_MAX_TURNS,
#         hooks: Optional[RunHooks[TContext]] = None,
#         run_config: Optional[RunConfig] = None,
#         previous_response_id: Optional[str] = None,
#         conversation_id: Optional[str] = None,
#         session: Optional[Session] = None,
#         metrics: Optional[List[BaseMetric]] = None,
#         metric_collection: Optional[str] = None,
#         name: Optional[str] = None,
#         tags: Optional[List[str]] = None,
#         metadata: Optional[dict] = None,
#         thread_id: Optional[str] = None,
#         user_id: Optional[str] = None,
#         **kwargs,  # backwards compatibility
#     ) -> RunResult:
#         is_agents_available()
#         # _patch_default_agent_runner_get_model()

#         with Observer(
#             span_type="custom",
#             metric_collection=metric_collection,
#             metrics=metrics,
#             func_name="run",
#             function_kwargs={"input": input},  # also set below
#         ) as observer:
#             update_trace_attributes(
#                 name=name,
#                 tags=tags,
#                 metadata=metadata,
#                 thread_id=thread_id,
#                 user_id=user_id,
#                 metric_collection=metric_collection,
#                 metrics=metrics,
#             )
#             current_span = current_span_context.get()
#             current_trace = current_trace_context.get()
#             if not current_trace.input:
#                 current_trace.input = input
#             if current_span:
#                 current_span.input = input
#             res = await super().run(
#                 starting_agent,
#                 input,
#                 context=context,
#                 max_turns=max_turns,
#                 hooks=hooks,
#                 run_config=run_config,
#                 previous_response_id=previous_response_id,
#                 conversation_id=conversation_id,
#                 session=session,
#                 **kwargs,  # backwards compatibility
#             )
#             current_trace_thread_id = current_trace_context.get().thread_id
#             _output = None
#             if current_trace_thread_id:
#                 _output = res.final_output
#             else:
#                 _output = str(res)
#             observer.result = _output
#             update_trace_attributes(output=_output)
#         return res

#     @classmethod
#     def run_sync(
#         cls,
#         starting_agent: Agent[TContext],
#         input: Union[str, list[TResponseInputItem]],
#         *,
#         context: Optional[TContext] = None,
#         max_turns: int = DEFAULT_MAX_TURNS,
#         hooks: Optional[RunHooks[TContext]] = None,
#         run_config: Optional[RunConfig] = None,
#         previous_response_id: Optional[str] = None,
#         conversation_id: Optional[str] = None,
#         session: Optional[Session] = None,
#         metrics: Optional[List[BaseMetric]] = None,
#         metric_collection: Optional[str] = None,
#         name: Optional[str] = None,
#         tags: Optional[List[str]] = None,
#         metadata: Optional[dict] = None,
#         thread_id: Optional[str] = None,
#         user_id: Optional[str] = None,
#         **kwargs,
#     ) -> RunResult:
#         is_agents_available()

#         with Observer(
#             span_type="custom",
#             metric_collection=metric_collection,
#             metrics=metrics,
#             func_name="run_sync",
#             function_kwargs={"input": input},  # also set below
#         ) as observer:
#             update_trace_attributes(
#                 name=name,
#                 tags=tags,
#                 metadata=metadata,
#                 thread_id=thread_id,
#                 user_id=user_id,
#                 metric_collection=metric_collection,
#                 metrics=metrics,
#             )

#             current_span = current_span_context.get()
#             current_trace = current_trace_context.get()
#             if not current_trace.input:
#                 current_trace.input = input
#             if current_span:
#                 current_span.input = input
#             res = super().run_sync(
#                 starting_agent,
#                 input,
#                 context=context,
#                 max_turns=max_turns,
#                 hooks=hooks,
#                 run_config=run_config,
#                 previous_response_id=previous_response_id,
#                 conversation_id=conversation_id,
#                 session=session,
#                 **kwargs,  # backwards compatibility
#             )
#             current_trace_thread_id = current_trace_context.get().thread_id
#             _output = None
#             if current_trace_thread_id:
#                 _output = res.final_output
#             else:
#                 _output = str(res)
#             update_trace_attributes(output=_output)
#             observer.result = _output

#         return res

#     @classmethod
#     def run_streamed(
#         cls,
#         starting_agent: Agent[TContext],
#         input: Union[str, list[TResponseInputItem]],
#         *,
#         context: Optional[TContext] = None,
#         max_turns: int = DEFAULT_MAX_TURNS,
#         hooks: Optional[RunHooks[TContext]] = None,
#         run_config: Optional[RunConfig] = None,
#         previous_response_id: Optional[str] = None,
#         conversation_id: Optional[str] = None,
#         session: Optional[Session] = None,
#         metrics: Optional[List[BaseMetric]] = None,
#         metric_collection: Optional[str] = None,
#         name: Optional[str] = None,
#         tags: Optional[List[str]] = None,
#         metadata: Optional[dict] = None,
#         thread_id: Optional[str] = None,
#         user_id: Optional[str] = None,
#         **kwargs,  # backwards compatibility
#     ) -> RunResultStreaming:
#         is_agents_available()
#         # Manually enter observer; we'll exit when streaming finishes
#         observer = Observer(
#             span_type="custom",
#             metric_collection=metric_collection,
#             metrics=metrics,
#             func_name="run_streamed",
#             function_kwargs={"input": input},
#         )
#         observer.__enter__()

#         update_trace_attributes(
#             name=name,
#             tags=tags,
#             metadata=metadata,
#             thread_id=thread_id,
#             user_id=user_id,
#             metric_collection=metric_collection,
#             metrics=metrics,
#         )
#         current_trace = current_trace_context.get()
#         if not current_trace.input:
#             current_trace.input = input

#         current_span = current_span_context.get()
#         if current_span:
#             current_span.input = input

#         res = super().run_streamed(
#             starting_agent,
#             input,
#             context=context,
#             max_turns=max_turns,
#             hooks=hooks,
#             run_config=run_config,
#             previous_response_id=previous_response_id,
#             conversation_id=conversation_id,
#             session=session,
#             **kwargs,  # backwards compatibility
#         )

#         # Runtime-patch stream_events so the observer closes only after streaming completes
#         orig_stream_events = res.stream_events

#         async def _patched_stream_events(self: RunResultStreaming):
#             try:
#                 async for event in orig_stream_events():
#                     yield event
#                 observer.result = self.final_output
#                 update_trace_attributes(output=self.final_output)
#             except Exception as e:
#                 observer.__exit__(type(e), e, e.__traceback__)
#                 raise
#             finally:
#                 observer.__exit__(None, None, None)

#         from types import MethodType as _MethodType

#         res.stream_events = _MethodType(_patched_stream_events, res)

#         return res


# def update_trace_attributes(
#     input: Any = None,
#     output: Any = None,
#     name: str = None,
#     tags: List[str] = None,
#     metadata: dict = None,
#     thread_id: str = None,
#     user_id: str = None,
#     metric_collection: str = None,
#     metrics: List[BaseMetric] = None,
# ):
#     current_trace = current_trace_context.get()
#     if input:
#         current_trace.input = input
#     if output:
#         current_trace.output = output
#     if name:
#         current_trace.name = name
#     if tags:
#         current_trace.tags = tags
#     if metadata:
#         current_trace.metadata = metadata
#     if thread_id:
#         current_trace.thread_id = thread_id
#     if user_id:
#         current_trace.user_id = user_id
#     if metric_collection:
#         current_trace.metric_collection = metric_collection
#     if metrics:
#         current_trace.metrics = metrics



import asyncio

from typing import Any
from deepeval.dataset.utils import coerce_to_task


class EvaluationTasks:

    def __init__(self):
        self._tasks: list[asyncio.Future] = []

    def append(self, obj: Any):
        self._tasks.append(coerce_to_task(obj))

    def get_tasks(self) -> list[asyncio.Future]:
        return list(self._tasks)

    def num_tasks(self):
        return len(self._tasks)

    def clear_tasks(self) -> None:
        for t in self._tasks:
            if not t.done():
                t.cancel()
        self._tasks.clear()



from typing import Optional, List, Tuple, Union, Dict

from deepeval.utils import get_or_create_event_loop
from deepeval.metrics.utils import (
    construct_verbose_logs,
    check_llm_test_case_params,
    initialize_model,
    a_generate_with_schema_and_extract,
    generate_with_schema_and_extract,
)
from deepeval.test_case import (
    LLMTestCase,
    LLMTestCaseParams,
)
from deepeval.metrics import BaseMetric
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.task_completion.template import TaskCompletionTemplate
from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.metrics.task_completion.schema import (
    TaskAndOutcome,
    TaskCompletionVerdict,
)
from deepeval.metrics.api import metric_data_manager


class TaskCompletionMetric(BaseMetric):

    _required_params: List[LLMTestCaseParams] = [
        LLMTestCaseParams.INPUT,
        LLMTestCaseParams.ACTUAL_OUTPUT,
    ]

    def __init__(
        self,
        threshold: float = 0.5,
        task: Optional[str] = None,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
    ):
        if task is None:
            self._is_task_provided = False
        else:
            self._is_task_provided = True

        self.task = task
        self.threshold = 1 if strict_mode else threshold
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.async_mode = async_mode
        self.strict_mode = strict_mode
        self.verbose_mode = verbose_mode
        self.requires_trace = True

    def measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        check_llm_test_case_params(
            test_case,
            self._required_params,
            None,
            None,
            self,
            self.model,
            test_case.multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                        _in_component=_in_component,
                        _log_metric_to_confident=_log_metric_to_confident,
                    )
                )
            else:
                task, self.outcome = self._extract_task_and_outcome(test_case)
                if self.task is None or not self._is_task_provided:
                    self.task = task
                self.verdict, self.reason = self._generate_verdicts()
                self.score = self._calculate_score()
                self.success = self.score >= self.threshold
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Task: {self.task}",
                        f"Outcome: {self.outcome}",
                        f"Score: {self.score}\nReason: {self.reason}",
                    ],
                )

                if _log_metric_to_confident:
                    metric_data_manager.post_metric_if_enabled(
                        self, test_case=test_case
                    )

            return self.score

    async def a_measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        check_llm_test_case_params(
            test_case,
            self._required_params,
            None,
            None,
            self,
            self.model,
            test_case.multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self,
            async_mode=True,
            _show_indicator=_show_indicator,
            _in_component=_in_component,
        ):
            task, self.outcome = await self._a_extract_task_and_outcome(
                test_case
            )
            if self.task is None or not self._is_task_provided:
                self.task = task
            self.verdict, self.reason = await self._a_generate_verdicts()
            self.score = self._calculate_score()
            self.success = self.score >= self.threshold
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Task: {self.task}",
                    f"Outcome: {self.outcome}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )

            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )

            return self.score

    async def _a_generate_verdicts(self) -> Tuple:
        prompt = TaskCompletionTemplate.generate_verdict(
            task=self.task,
            actual_outcome=self.outcome,
        )
        return await a_generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=TaskCompletionVerdict,
            extract_schema=lambda s: (s.verdict, s.reason),
            extract_json=lambda data: (data["verdict"], data["reason"]),
        )

    def _generate_verdicts(self) -> Tuple:
        prompt = TaskCompletionTemplate.generate_verdict(
            task=self.task,
            actual_outcome=self.outcome,
        )
        return generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=TaskCompletionVerdict,
            extract_schema=lambda s: (s.verdict, s.reason),
            extract_json=lambda data: (data["verdict"], data["reason"]),
        )

    async def _a_extract_task_and_outcome(
        self,
        test_case: LLMTestCase,
    ) -> Tuple:
        has_trace: bool = isinstance(test_case._trace_dict, Dict)
        if has_trace:
            prompt = TaskCompletionTemplate.extract_task_and_outcome_from_trace(
                trace=test_case._trace_dict
            )
        else:
            # TODO: Deprecate this soon
            prompt = TaskCompletionTemplate.extract_goal_and_outcome(
                input=test_case.input,
                actual_output=test_case.actual_output,
                tools_called=test_case.tools_called,
            )
        return await a_generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=TaskAndOutcome,
            extract_schema=lambda s: (s.task, s.outcome),
            extract_json=lambda data: (data["task"], data["outcome"]),
        )

    def _extract_task_and_outcome(
        self,
        test_case: LLMTestCase,
    ) -> Tuple:
        has_trace: bool = isinstance(test_case._trace_dict, Dict)
        if has_trace:
            prompt = TaskCompletionTemplate.extract_task_and_outcome_from_trace(
                trace=test_case._trace_dict
            )
        else:
            # TODO: Deprecate this soon
            prompt = TaskCompletionTemplate.extract_goal_and_outcome(
                input=test_case.input,
                actual_output=test_case.actual_output,
                tools_called=test_case.tools_called,
            )
        return generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=TaskAndOutcome,
            extract_schema=lambda s: (s.task, s.outcome),
            extract_json=lambda data: (data["task"], data["outcome"]),
        )

    def _calculate_score(self):
        return (
            0
            if self.strict_mode and self.verdict < self.threshold
            else self.verdict
        )

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except TypeError:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "Task Completion"



from typing import List, Dict, Any, Sequence
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter
from opentelemetry.sdk.trace.export import SpanExportResult
import json
from datetime import datetime


class TestExporter(SpanExporter):
    """This exporter is used to test the exporter. It will store the spans in a list of dictionaries."""

    span_json_list: List[Dict[str, Any]] = []

    def export(
        self, spans: Sequence[ReadableSpan], timeout_millis: int = 30000
    ) -> SpanExportResult:
        for span in spans:
            _span_json = json.loads(span.to_json())
            self.span_json_list.append(_span_json)

        return SpanExportResult.SUCCESS

    def get_span_json_list(self) -> List[Dict[str, Any]]:
        return sorted(
            self.span_json_list,
            key=lambda x: datetime.fromisoformat(
                x["start_time"].replace("Z", "+00:00")
            ),
        )

    def clear_span_json_list(self):
        self.span_json_list = []


test_exporter = TestExporter()



"""A slightly modified tailored version of the LLM evaluated metric based on the GEval framework: https://arxiv.org/pdf/2303.16634.pdf"""

from openai.types.chat.chat_completion import ChatCompletion
from typing import Optional, List, Tuple, Union, Dict, Type
import math
from deepeval.metrics import BaseConversationalMetric
from deepeval.metrics.g_eval.utils import (
    Rubric,
    construct_conversational_g_eval_turn_params_string,
    construct_non_turns_test_case_string,
    format_rubrics,
    validate_and_sort_rubrics,
    validate_criteria_and_evaluation_steps,
)
from deepeval.test_case import (
    TurnParams,
    ConversationalTestCase,
)
from deepeval.metrics.conversational_g_eval.template import (
    ConversationalGEvalTemplate,
)
from deepeval.utils import get_or_create_event_loop, prettify_list
from deepeval.metrics.utils import (
    check_conversational_test_case_params,
    construct_verbose_logs,
    trimAndLoadJson,
    initialize_model,
    convert_turn_to_dict,
    a_generate_with_schema_and_extract,
    generate_with_schema_and_extract,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.indicator import metric_progress_indicator
import deepeval.metrics.conversational_g_eval.schema as cgschema
from deepeval.metrics.api import metric_data_manager


class ConversationalGEval(BaseConversationalMetric):
    def __init__(
        self,
        name: str,
        evaluation_params: Optional[List[TurnParams]] = None,
        criteria: Optional[str] = None,
        evaluation_steps: Optional[List[str]] = None,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        threshold: float = 0.5,
        rubric: Optional[List[Rubric]] = None,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
        evaluation_template: Type[
            ConversationalGEvalTemplate
        ] = ConversationalGEvalTemplate,
        _include_g_eval_suffix: bool = True,
    ):
        if evaluation_params is not None and len(evaluation_params) == 0:
            raise ValueError("evaluation_params cannot be an empty list.")

        self.name = name
        if evaluation_params is None:
            evaluation_params = [TurnParams.CONTENT, TurnParams.ROLE]

        if TurnParams.CONTENT not in evaluation_params:
            evaluation_params.append(TurnParams.CONTENT)
        if TurnParams.ROLE not in evaluation_params:
            evaluation_params.append(TurnParams.ROLE)

        self.evaluation_params = evaluation_params

        validate_criteria_and_evaluation_steps(criteria, evaluation_steps)
        self.criteria = criteria
        self.rubric = validate_and_sort_rubrics(rubric)
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.evaluation_steps = (
            evaluation_steps
            if evaluation_steps and len(evaluation_steps) > 0
            else None
        )
        self.threshold = 1 if strict_mode else threshold
        self.strict_mode = strict_mode
        self.async_mode = async_mode
        self.verbose_mode = verbose_mode
        self.evaluation_template = evaluation_template
        self._include_g_eval_suffix = _include_g_eval_suffix

    def measure(
        self,
        test_case: ConversationalTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        multimodal = test_case.multimodal
        check_conversational_test_case_params(
            test_case,
            self.evaluation_params,
            self,
            False,
            self.model,
            multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                        _in_component=_in_component,
                        _log_metric_to_confident=_log_metric_to_confident,
                    )
                )
            else:
                self.evaluation_steps: List[str] = (
                    self._generate_evaluation_steps()
                )
                g_score, reason = self.evaluate(test_case)
                self.reason = reason
                self.score = float(g_score) / 10
                self.score = (
                    0
                    if self.strict_mode and self.score < self.threshold
                    else self.score
                )
                self.success = self.score >= self.threshold
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Criteria:\n{self.criteria}",
                        f"Evaluation Steps:\n{prettify_list(self.evaluation_steps)}",
                        f"Rubric:\n{format_rubrics(self.rubric)}",
                        f"Score: {self.score}\nReason: {self.reason}",
                    ],
                )
                if _log_metric_to_confident:
                    metric_data_manager.post_metric_if_enabled(
                        self, test_case=test_case
                    )

            return self.score

    async def a_measure(
        self,
        test_case: ConversationalTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        multimodal = test_case.multimodal
        check_conversational_test_case_params(
            test_case,
            self.evaluation_params,
            self,
            False,
            self.model,
            multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self,
            async_mode=True,
            _show_indicator=_show_indicator,
            _in_component=_in_component,
        ):
            self.evaluation_steps: List[str] = (
                await self._a_generate_evaluation_steps()
            )
            g_score, reason = await self._a_evaluate(test_case)
            self.reason = reason
            self.score = float(g_score) / 10
            self.score = (
                0
                if self.strict_mode and self.score < self.threshold
                else self.score
            )
            self.success = self.score >= self.threshold
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Criteria:\n{self.criteria}",
                    f"Evaluation Steps:\n{prettify_list(self.evaluation_steps)}",
                    f"Rubric:\n{format_rubrics(self.rubric)}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )
            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )

            return self.score

    async def _a_generate_evaluation_steps(self) -> List[str]:
        if self.evaluation_steps:
            return self.evaluation_steps

        g_eval_params_str = construct_conversational_g_eval_turn_params_string(
            self.evaluation_params
        )
        prompt = self.evaluation_template.generate_evaluation_steps(
            criteria=self.criteria, parameters=g_eval_params_str
        )
        return await a_generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=cgschema.Steps,
            extract_schema=lambda s: s.steps,
            extract_json=lambda data: data["steps"],
        )

    def _generate_evaluation_steps(self) -> List[str]:
        if self.evaluation_steps:
            return self.evaluation_steps

        g_eval_params_str = construct_conversational_g_eval_turn_params_string(
            self.evaluation_params
        )
        prompt = self.evaluation_template.generate_evaluation_steps(
            criteria=self.criteria, parameters=g_eval_params_str
        )
        return generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=cgschema.Steps,
            extract_schema=lambda s: s.steps,
            extract_json=lambda data: data["steps"],
        )

    async def _a_evaluate(
        self, test_case: ConversationalTestCase
    ) -> Tuple[Union[int, float], str]:
        test_case_content = construct_non_turns_test_case_string(
            self.evaluation_params, test_case
        )
        g_eval_params_str = construct_conversational_g_eval_turn_params_string(
            self.evaluation_params
        )
        if not self.strict_mode:
            rubric_str = format_rubrics(self.rubric) if self.rubric else None
            prompt = self.evaluation_template.generate_evaluation_results(
                evaluation_steps=self.number_evaluation_steps(),
                test_case_content=test_case_content,
                turns=[
                    convert_turn_to_dict(turn, self.evaluation_params)
                    for turn in test_case.turns
                ],
                parameters=g_eval_params_str,
                rubric=rubric_str,
            )
        else:
            prompt = self.evaluation_template.generate_evaluation_results(
                evaluation_steps=self.number_evaluation_steps(),
                test_case_content=test_case_content,
                turns=[
                    convert_turn_to_dict(turn, self.evaluation_params)
                    for turn in test_case.turns
                ],
                parameters=g_eval_params_str,
            )
        try:
            res, cost = await self.model.a_generate_raw_response(
                prompt, top_logprobs=20
            )

            self._accrue_cost(cost)
            data = trimAndLoadJson(res.choices[0].message.content, self)

            reason = data["reason"]
            score = data["score"]
            if self.strict_mode:
                return score, reason

            try:
                weighted_summed_score = self.generate_weighted_summed_score(
                    score, res
                )
                return weighted_summed_score, reason
            except (KeyError, AttributeError, TypeError, ValueError):
                return score, reason
        except (
            AttributeError
        ):  # This catches the case where a_generate_raw_response doesn't exist.
            return await a_generate_with_schema_and_extract(
                metric=self,
                prompt=prompt,
                schema_cls=cgschema.ReasonScore,
                extract_schema=lambda r: (r.score, r.reason),
                extract_json=lambda data: (data["score"], data["reason"]),
            )

    def evaluate(
        self, test_case: ConversationalTestCase
    ) -> Tuple[Union[int, float], str]:
        test_case_content = construct_non_turns_test_case_string(
            self.evaluation_params, test_case
        )
        g_eval_params_str = construct_conversational_g_eval_turn_params_string(
            self.evaluation_params
        )
        if not self.strict_mode:
            rubric_str = format_rubrics(self.rubric) if self.rubric else None
            prompt = self.evaluation_template.generate_evaluation_results(
                evaluation_steps=self.number_evaluation_steps(),
                test_case_content=test_case_content,
                turns=[
                    convert_turn_to_dict(turn, self.evaluation_params)
                    for turn in test_case.turns
                ],
                parameters=g_eval_params_str,
                rubric=rubric_str,
            )
        else:
            prompt = self.evaluation_template.generate_evaluation_results(
                evaluation_steps=self.number_evaluation_steps(),
                test_case_content=test_case_content,
                turns=[
                    convert_turn_to_dict(turn, self.evaluation_params)
                    for turn in test_case.turns
                ],
                parameters=g_eval_params_str,
            )
        try:
            res, cost = self.model.generate_raw_response(
                prompt, top_logprobs=20
            )
            self._accrue_cost(cost)
            data = trimAndLoadJson(res.choices[0].message.content, self)

            reason = data["reason"]
            score = data["score"]
            if self.strict_mode:
                return score, reason

            try:
                weighted_summed_score = self.generate_weighted_summed_score(
                    score, res
                )
                return weighted_summed_score, reason
            except (KeyError, AttributeError, TypeError, ValueError):
                return score, reason
        except AttributeError:
            # This catches the case where a_generate_raw_response doesn't exist.
            return generate_with_schema_and_extract(
                metric=self,
                prompt=prompt,
                schema_cls=cgschema.ReasonScore,
                extract_schema=lambda r: (r.score, r.reason),
                extract_json=lambda data: (data["score"], data["reason"]),
            )

    def generate_weighted_summed_score(
        self, raw_score: int, raw_response: ChatCompletion
    ) -> Union[int, float]:
        generated_logprobs = raw_response.choices[0].logprobs.content
        # First, locate the token that we care for logprobs, i.e., the token matching the score
        score_logprobs = None
        for token_logprobs in generated_logprobs:
            if token_logprobs.token == str(raw_score):
                score_logprobs = token_logprobs
                break
        # Then, calculate the score based on the logprobs
        token_linear_probability: Dict[int, float] = {}
        sum_linear_probability = 0
        # Filter out tokens with <1% linear probability, i.e., logprobs < math.log(0.01)
        min_logprob = math.log(0.01)
        for token_logprob in score_logprobs.top_logprobs:
            logprob = token_logprob.logprob

            # Filter out low probability tokens
            if logprob < min_logprob:
                continue
            # Filter out non-decimal token to prevent errors in later int(token) conversion
            if not token_logprob.token.isdecimal():
                continue

            # Calculate the linear probability
            linear_prob = math.exp(logprob)
            token_score = int(token_logprob.token)
            if token_linear_probability.get(token_score):
                token_linear_probability[token_score] += linear_prob
            else:
                token_linear_probability[token_score] = linear_prob
            sum_linear_probability += linear_prob

        sum_of_weighted_scores = 0.0
        for score, prob in token_linear_probability.items():
            sum_of_weighted_scores += score * prob

        # Scale the sum of linear probability to 1
        weighted_summed_score = sum_of_weighted_scores / sum_linear_probability
        return weighted_summed_score

    def number_evaluation_steps(self):
        evaluation_steps = """"""
        for index, string in enumerate(self.evaluation_steps, start=1):
            evaluation_steps += f"{index}. {string}\n"
        return evaluation_steps

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except TypeError:
                self.success = False
        return self.success

    @property
    def __name__(self):
        if self._include_g_eval_suffix:
            return f"{self.name} [Conversational GEval]"
        else:
            return self.name



from typing import List
from textwrap import dedent


multimodal_rules = """
    --- MULTIMODAL INPUT RULES ---
    - Treat image content as factual evidence.
    - Only reference visual details that are explicitly and clearly visible.
    - Do not infer or guess objects, text, or details not visibly present.
    - If an image is unclear or ambiguous, mark uncertainty explicitly.
"""


class ConversationalVerdictNodeTemplate:

    @staticmethod
    def generate_reason(verbose_steps: List[str], score: float, name: str):
        return dedent(
            f"""You are given a metric name, its score, and a traversal path through a conversational evaluation DAG (Directed Acyclic Graph).
                This DAG reflects step-by-step reasoning over a dialogue to arrive at the final verdict.

                Each step in the DAG represents a judgment based on parts of the conversation — including roles and the contents they spoke of.

                Your task is to explain **why the score was assigned**, using the traversal steps to justify the reasoning.

                Metric Name:
                {name}

                Score:
                {score}

                DAG Traversal:
                {verbose_steps}

                **
                IMPORTANT: Only return JSON with a 'reason' key.
                Example:
                {{
                "reason": "The score is {score} because the assistant repeatedly failed to clarify the user's ambiguous statements, as shown in the DAG traversal path."
                }}
                **
                JSON:
            """
        )


class ConversationalTaskNodeTemplate:
    @staticmethod
    def generate_task_output(instructions: str, text: str):
        return dedent(
            f"""You are given a set of task instructions and a full conversation between a user and an assistant.

                {multimodal_rules}

                Instructions:
                {instructions}

                {text}

                ===END OF INPUT===

                **
                IMPORTANT: Only return a JSON with the 'output' key containing the result of applying the instructions to the conversation.
                Example:
                {{
                "output": "..."
                }}
                **
                JSON:
            """
        )


class ConversationalBinaryJudgementTemplate:
    @staticmethod
    def generate_binary_verdict(criteria: str, text: str):
        return dedent(
            f"""{criteria}

                Below is the full conversation you should evaluate. Consider dialogue context, speaker roles, and how responses were handled.

                {multimodal_rules}

                Full Conversation:
                {text}

                **
                IMPORTANT: Only return JSON with two keys:
                - 'verdict': true or false
                - 'reason': justification based on specific parts of the conversation

                Example:
                {{
                "reason": "The assistant provided a clear and direct answer in response to every user query.",
                "verdict": true
                }}
                **
                JSON:
            """
        )


class ConversationalNonBinaryJudgementTemplate:
    @staticmethod
    def generate_non_binary_verdict(
        criteria: str, text: str, options: List[str]
    ):
        return dedent(
            f"""{criteria}

                You are evaluating the following conversation. Choose one of the options that best reflects the assistant's behavior.

                {multimodal_rules}

                Options: {options}

                Full Conversation:
                {text}

                **
                IMPORTANT: Only return JSON with two keys:
                - 'verdict': one of the listed options
                - 'reason': explanation referencing specific conversation points

                Example:
                {{
                "reason": "The assistant partially addressed the user's issue but missed clarifying their follow-up question.",
                "verdict": "{options[1]}"
                }}
                **
                JSON:
            """
        )



from typing import Optional, List, Type, Union

from deepeval.utils import (
    get_or_create_event_loop,
    prettify_list,
)
from deepeval.metrics.utils import (
    construct_verbose_logs,
    check_llm_test_case_params,
    initialize_model,
    a_generate_with_schema_and_extract,
    generate_with_schema_and_extract,
)
from deepeval.test_case import (
    LLMTestCase,
    LLMTestCaseParams,
)
from deepeval.metrics import BaseMetric
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.contextual_recall.template import ContextualRecallTemplate
from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.metrics.contextual_recall.schema import (
    ContextualRecallVerdict,
    Verdicts,
    ContextualRecallScoreReason,
)
from deepeval.metrics.api import metric_data_manager


class ContextualRecallMetric(BaseMetric):

    _required_params: List[LLMTestCaseParams] = [
        LLMTestCaseParams.INPUT,
        LLMTestCaseParams.RETRIEVAL_CONTEXT,
        LLMTestCaseParams.EXPECTED_OUTPUT,
    ]

    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
        evaluation_template: Type[
            ContextualRecallTemplate
        ] = ContextualRecallTemplate,
    ):
        self.threshold = 1 if strict_mode else threshold
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.async_mode = async_mode
        self.strict_mode = strict_mode
        self.verbose_mode = verbose_mode
        self.evaluation_template = evaluation_template

    def measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        multimodal = test_case.multimodal

        check_llm_test_case_params(
            test_case,
            self._required_params,
            None,
            None,
            self,
            self.model,
            test_case.multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                        _in_component=_in_component,
                        _log_metric_to_confident=_log_metric_to_confident,
                    )
                )
            else:
                expected_output = test_case.expected_output
                retrieval_context = test_case.retrieval_context

                self.verdicts: List[ContextualRecallVerdict] = (
                    self._generate_verdicts(
                        expected_output, retrieval_context, multimodal
                    )
                )
                self.score = self._calculate_score()
                self.reason = self._generate_reason(expected_output, multimodal)
                self.success = self.score >= self.threshold
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Verdicts:\n{prettify_list(self.verdicts)}",
                        f"Score: {self.score}\nReason: {self.reason}",
                    ],
                )
                if _log_metric_to_confident:
                    metric_data_manager.post_metric_if_enabled(
                        self, test_case=test_case
                    )
            return self.score

    async def a_measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:

        multimodal = test_case.multimodal

        check_llm_test_case_params(
            test_case,
            self._required_params,
            None,
            None,
            self,
            self.model,
            test_case.multimodal,
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self,
            async_mode=True,
            _show_indicator=_show_indicator,
            _in_component=_in_component,
        ):
            expected_output = test_case.expected_output
            retrieval_context = test_case.retrieval_context

            self.verdicts: List[ContextualRecallVerdict] = (
                await self._a_generate_verdicts(
                    expected_output, retrieval_context, multimodal
                )
            )
            self.score = self._calculate_score()
            self.reason = await self._a_generate_reason(
                expected_output, multimodal
            )
            self.success = self.score >= self.threshold
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Verdicts:\n{prettify_list(self.verdicts)}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )
            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )
            return self.score

    async def _a_generate_reason(self, expected_output: str, multimodal: bool):
        if self.include_reason is False:
            return None

        supportive_reasons = []
        unsupportive_reasons = []
        for verdict in self.verdicts:
            if verdict.verdict.lower() == "yes":
                supportive_reasons.append(verdict.reason)
            else:
                unsupportive_reasons.append(verdict.reason)

        prompt = self.evaluation_template.generate_reason(
            expected_output=expected_output,
            supportive_reasons=supportive_reasons,
            unsupportive_reasons=unsupportive_reasons,
            score=format(self.score, ".2f"),
            multimodal=multimodal,
        )

        return await a_generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=ContextualRecallScoreReason,
            extract_schema=lambda score_reason: score_reason.reason,
            extract_json=lambda data: data["reason"],
        )

    def _generate_reason(self, expected_output: str, multimodal: bool):
        if self.include_reason is False:
            return None

        supportive_reasons = []
        unsupportive_reasons = []
        for verdict in self.verdicts:
            if verdict.verdict.lower() == "yes":
                supportive_reasons.append(verdict.reason)
            else:
                unsupportive_reasons.append(verdict.reason)

        prompt = self.evaluation_template.generate_reason(
            expected_output=expected_output,
            supportive_reasons=supportive_reasons,
            unsupportive_reasons=unsupportive_reasons,
            score=format(self.score, ".2f"),
            multimodal=multimodal,
        )

        return generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=ContextualRecallScoreReason,
            extract_schema=lambda score_reason: score_reason.reason,
            extract_json=lambda data: data["reason"],
        )

    def _calculate_score(self):
        number_of_verdicts = len(self.verdicts)
        if number_of_verdicts == 0:
            return 0

        justified_sentences = 0
        for verdict in self.verdicts:
            if verdict.verdict.lower() == "yes":
                justified_sentences += 1

        score = justified_sentences / number_of_verdicts
        return 0 if self.strict_mode and score < self.threshold else score

    async def _a_generate_verdicts(
        self,
        expected_output: str,
        retrieval_context: List[str],
        multimodal: bool,
    ) -> List[ContextualRecallVerdict]:
        prompt = self.evaluation_template.generate_verdicts(
            expected_output=expected_output,
            retrieval_context=retrieval_context,
            multimodal=multimodal,
        )
        return await a_generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=Verdicts,
            extract_schema=lambda r: list(r.verdicts),
            extract_json=lambda data: [
                ContextualRecallVerdict(**item) for item in data["verdicts"]
            ],
        )

    def _generate_verdicts(
        self,
        expected_output: str,
        retrieval_context: List[str],
        multimodal: bool,
    ) -> List[ContextualRecallVerdict]:
        prompt = self.evaluation_template.generate_verdicts(
            expected_output=expected_output,
            retrieval_context=retrieval_context,
            multimodal=multimodal,
        )
        return generate_with_schema_and_extract(
            metric=self,
            prompt=prompt,
            schema_cls=Verdicts,
            extract_schema=lambda r: list(r.verdicts),
            extract_json=lambda data: [
                ContextualRecallVerdict(**item) for item in data["verdicts"]
            ],
        )

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except TypeError:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "Contextual Recall"



from typing import List

from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.metrics.utils import (
    check_llm_test_case_params,
    construct_verbose_logs,
)
from deepeval.metrics.api import metric_data_manager
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase, LLMTestCaseParams


class ExactMatchMetric(BaseMetric):
    _required_params: List[LLMTestCaseParams] = [
        LLMTestCaseParams.INPUT,
        LLMTestCaseParams.ACTUAL_OUTPUT,
        LLMTestCaseParams.EXPECTED_OUTPUT,
    ]

    def __init__(
        self,
        threshold: float = 1,
        verbose_mode: bool = False,
    ):
        self.threshold = threshold
        self.verbose_mode = verbose_mode

    def measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        check_llm_test_case_params(
            test_case,
            self._required_params,
            None,
            None,
            self,
            None,
            test_case.multimodal,
        )

        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            expected = test_case.expected_output.strip()
            actual = test_case.actual_output.strip()

            if expected == actual:
                self.score = self.precision = self.recall = self.f1 = 1.0
                self.reason = (
                    "The actual and expected outputs are exact matches."
                )
            else:
                self.score = self.precision = self.recall = self.f1 = 0.0
                self.reason = "The actual and expected outputs are different."

            self.success = self.score >= self.threshold

            if self.verbose_mode:
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Score: {self.score:.2f}",
                        f"Reason: {self.reason}",
                    ],
                )

            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )

            return self.score

    async def a_measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
    ) -> float:
        return self.measure(
            test_case,
            _show_indicator=_show_indicator,
            _in_component=_in_component,
        )

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "Exact Match"



from enum import Enum
from typing import Dict, List, Optional, Union, Literal, Any
from pydantic import BaseModel, Field

from deepeval.test_case import ToolCall
from deepeval.utils import make_model_config


class SpanApiType(Enum):
    BASE = "base"
    AGENT = "agent"
    LLM = "llm"
    RETRIEVER = "retriever"
    TOOL = "tool"


span_api_type_literals = Literal["base", "agent", "llm", "retriever", "tool"]


class TraceSpanApiStatus(Enum):
    SUCCESS = "SUCCESS"
    ERRORED = "ERRORED"


class PromptApi(BaseModel):
    alias: Optional[str] = None
    version: Optional[str] = None


class MetricData(BaseModel):
    model_config = make_model_config(extra="ignore")

    name: str
    threshold: float
    success: bool
    score: Optional[float] = None
    reason: Optional[str] = None
    strict_mode: Optional[bool] = Field(False, alias="strictMode")
    evaluation_model: Optional[str] = Field(None, alias="evaluationModel")
    error: Optional[str] = None
    evaluation_cost: Union[float, None] = Field(None, alias="evaluationCost")
    verbose_logs: Optional[str] = Field(None, alias="verboseLogs")


class BaseApiSpan(BaseModel):
    model_config = make_model_config(
        use_enum_values=True, validate_assignment=True
    )

    uuid: str
    name: str = None
    status: TraceSpanApiStatus
    type: SpanApiType
    parent_uuid: Optional[str] = Field(None, alias="parentUuid")
    start_time: str = Field(alias="startTime")
    end_time: str = Field(alias="endTime")
    metadata: Optional[Dict[str, Any]] = None
    input: Optional[Any] = Field(None)
    output: Optional[Any] = Field(None)
    error: Optional[str] = None

    # additional test case parameters
    retrieval_context: Optional[List[str]] = Field(
        None, alias="retrievalContext"
    )
    context: Optional[List[str]] = Field(None, alias="context")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    tools_called: Optional[List[ToolCall]] = Field(None, alias="toolsCalled")
    expected_tools: Optional[List[ToolCall]] = Field(
        None, alias="expectedTools"
    )

    # agents
    available_tools: Optional[List[str]] = Field(None, alias="availableTools")
    agent_handoffs: Optional[List[str]] = Field(None, alias="agentHandoffs")

    # tools
    description: Optional[str] = None

    # retriever
    embedder: Optional[str] = None
    top_k: Optional[int] = Field(None, alias="topK")
    chunk_size: Optional[int] = Field(None, alias="chunkSize")

    # llm
    model: Optional[str] = None
    prompt: Optional[PromptApi] = None
    input_token_count: Optional[float] = Field(None, alias="inputTokenCount")
    output_token_count: Optional[float] = Field(None, alias="outputTokenCount")
    cost_per_input_token: Optional[float] = Field(
        None, alias="costPerInputToken"
    )
    cost_per_output_token: Optional[float] = Field(
        None, alias="costPerOutputToken"
    )
    token_intervals: Optional[Dict[str, str]] = Field(
        None, alias="tokenIntervals"
    )

    ## evals
    metric_collection: Optional[str] = Field(None, alias="metricCollection")
    metrics_data: Optional[List[MetricData]] = Field(None, alias="metricsData")


class TraceApi(BaseModel):
    model_config = make_model_config(
        use_enum_values=True, validate_assignment=True
    )

    uuid: str
    base_spans: Optional[List[BaseApiSpan]] = Field(None, alias="baseSpans")
    agent_spans: Optional[List[BaseApiSpan]] = Field(None, alias="agentSpans")
    llm_spans: Optional[List[BaseApiSpan]] = Field(None, alias="llmSpans")
    retriever_spans: Optional[List[BaseApiSpan]] = Field(
        None, alias="retrieverSpans"
    )
    tool_spans: Optional[List[BaseApiSpan]] = Field(None, alias="toolSpans")
    start_time: str = Field(alias="startTime")
    end_time: str = Field(alias="endTime")
    name: Optional[str] = Field(None)
    metadata: Optional[Dict[str, Any]] = Field(None)
    tags: Optional[List[str]] = Field(None)
    environment: Optional[str] = Field(None)
    thread_id: Optional[str] = Field(None, alias="threadId")
    user_id: Optional[str] = Field(None, alias="userId")
    input: Optional[Any] = Field(None)
    output: Optional[Any] = Field(None)
    status: Optional[TraceSpanApiStatus] = Field(TraceSpanApiStatus.SUCCESS)

    # additional test case parameters
    retrieval_context: Optional[List[str]] = Field(
        None, alias="retrievalContext"
    )
    context: Optional[List[str]] = Field(None, alias="context")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    tools_called: Optional[List[ToolCall]] = Field(None, alias="toolsCalled")
    expected_tools: Optional[List[ToolCall]] = Field(
        None, alias="expectedTools"
    )

    # evals
    metric_collection: Optional[str] = Field(None, alias="metricCollection")
    metrics_data: Optional[List[MetricData]] = Field(None, alias="metricsData")

    # Don't serialize these
    confident_api_key: Optional[str] = Field(None, exclude=True)



try:
    import anthropic  # noqa: F401
except ImportError:
    raise ModuleNotFoundError(
        "Please install anthropic to use this feature: 'pip install anthropic'"
    )

try:
    from anthropic import Anthropic, AsyncAnthropic  # noqa: F401
except ImportError:
    Anthropic = None  # type: ignore
    AsyncAnthropic = None  # type: ignore

if Anthropic or AsyncAnthropic:
    from deepeval.anthropic.patch import patch_anthropic_classes
    from deepeval.telemetry import capture_tracing_integration

    with capture_tracing_integration("anthropic"):
        patch_anthropic_classes()



from pydantic import BaseModel
from typing import List


class GoalSteps(BaseModel):
    user_goal: str
    steps_taken: List[str]


class GoalScore(BaseModel):
    score: float
    reason: str


class PlanScore(BaseModel):
    score: float
    reason: str



import logging
from typing import Any, Dict, Optional, Tuple
from pydantic import SecretStr

from deepeval.errors import DeepEvalError


logger = logging.getLogger(__name__)


def parse_model_name(model_name: Optional[str] = None) -> Optional[str]:
    """Extract base model name from provider-prefixed format.

    This function is useful for extracting the actual model name from a
    provider-prefixed format which is used by some proxies like LiteLLM.
    LiteLLM is designed to work with many different LLM providers (OpenAI, Anthropic,
    Cohere, etc.). To tell it which provider's API to call, you prepend the provider
    name to the model ID, in the form "<provider>/<model>". So openai/gpt-4.1-mini
    literally means "OpenAI's GPT-4.1 Mini via the OpenAI chat completions endpoint."

    Args:
        model_name: Original model identifier, potentially in
            "<provider>/<model>" format

    Returns:
        The model name without provider prefix

    Examples:
        parse_model_name("openai/gpt-4o") -> "gpt-4o"
        parse_model_name("gpt-4o") -> "gpt-4o"
    """
    if model_name is None:
        return None

    # if "/" in model_name:
    #     _, parsed_model_name = model_name.split("/", 1)
    #     return parsed_model_name
    return model_name


def require_secret_api_key(
    secret: Optional[SecretStr],
    *,
    provider_label: str,
    env_var_name: str,
    param_hint: str,
) -> str:
    """
    Normalize and validate a provider API key stored as a SecretStr.

    Args:
        secret:
            The SecretStr coming from Settings or an explicit constructor arg.
        provider_label:
            Human readable provider name for error messages, such as Anthropic, or OpenAI etc
        env_var_name:
            The environment variable backing this key
        param_hint:
            A short hint telling users how to pass the key explicitly

    Returns:
        The underlying API key string.

    Raises:
        DeepEvalError: if the key is missing or empty.
    """
    if secret is None:
        raise DeepEvalError(
            f"{provider_label} API key is not configured. "
            f"Set {env_var_name} in your environment or pass "
            f"{param_hint}."
        )

    api_key = secret.get_secret_value()
    if not api_key:
        raise DeepEvalError(
            f"{provider_label} API key is empty. Please configure a valid key."
        )

    return api_key


def require_costs(
    model_data,
    model_name: str,
    input_token_envvar: str,
    output_token_envvar: str,
    cost_per_input_token: Optional[float] = None,
    cost_per_output_token: Optional[float] = None,
) -> Tuple[Optional[float], Optional[float]]:
    """
    Validates and returns the cost parameters (input and output tokens) for a model.

    Arguments:
    - model_data: The model's data object, which should contain `input_price` and `output_price`.
    - model_name: The model name used for error messaging.
    - cost_per_input_token: The input token cost provided during model initialization (optional).
    - cost_per_output_token: The output token cost provided during model initialization (optional).
    - input_token_envvar: The environment variable name for input cost.
    - output_token_envvar: The environment variable name for output cost.

    Returns:
    - A tuple of validated values (input_cost, output_cost). If the values are provided, they are returned.
      If not provided, they are fetched from settings or environment variables.
    """

    def validate_cost(
        value: Optional[float], envvar_name: str
    ) -> Optional[float]:
        """Helper function to validate the cost values."""
        if value is not None and value < 0:
            raise DeepEvalError(f"{envvar_name} must be >= 0.")
        return value

    # Validate provided token costs
    cost_per_input_token = validate_cost(
        cost_per_input_token, input_token_envvar
    )
    cost_per_output_token = validate_cost(
        cost_per_output_token, output_token_envvar
    )

    # If model data doesn't have pricing, use provided values or environment variables
    if model_data.input_price is None or model_data.output_price is None:
        if cost_per_input_token is None or cost_per_output_token is None:
            raise DeepEvalError(
                f"No pricing available for `{model_name}`. "
                f"Please provide both `cost_per_input_token` and `cost_per_output_token` when initializing `{model_name}`, "
                f"or set {input_token_envvar} and {output_token_envvar} environment variables."
            )

        # Return the validated cost values as a tuple
        return cost_per_input_token, cost_per_output_token

    # If no custom cost values are provided, return model's default cost values
    return model_data.input_price, model_data.output_price


def normalize_kwargs_and_extract_aliases(
    provider_label: str,
    kwargs: Dict[str, Any],
    alias_map: Dict[str, list],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Normalize legacy keyword argument names according to alias_map.

    alias_map is of the form: {new_name: [old_name1, old_name2, ...]}

    - Returns (normalized_kwargs, extracted_values)
      where:
        - normalized_kwargs has all legacy keys removed (to prevent forwarding
          to downstream SDK clients).
        - extracted_values maps new_name -> value for any alias that was used.

    - Logs a warning for each legacy keyword used, so callers know they should
      migrate to the new name.
    """
    normalized = dict(kwargs)
    extracted: Dict[str, Any] = {}

    for new_name, old_names in alias_map.items():
        for old_name in old_names:
            if old_name in normalized:
                value = normalized.pop(old_name)

                logger.warning(
                    "%s keyword '%s' is deprecated; please use '%s' instead.",
                    provider_label,
                    old_name,
                    new_name,
                )

                # Only preserve the first alias value we see for a given new_name
                if new_name not in extracted:
                    extracted[new_name] = value

    return normalized, extracted



# - SIMBA-style multi-strategy 0-shot variant:
#   - Works on a single set of goldens (no D_pareto split).
#   - Maintains a bounded population of candidate prompts
#     (size controlled by `population_size`).
#   - At each iteration:
#       - Select a parent via epsilon-greedy on mean minibatch score.
#       - Sample a minibatch of goldens for scoring.
#       - Compute feedback once for the parent + minibatch.
#       - Propose multiple child prompts cooperatively from the same parent
#         (up to `proposals_per_step` children), each using a SIMBA edit
#         strategy (e.g., APPEND_DEMO or APPEND_RULE).
#       - For each child, accept it if its minibatch score improves on the
#         parent by at least `min_delta`, add it to the pool, and prune
#         low-scoring candidates if the population exceeds `population_size`.
#   - Uses `full_eval_every` (if set) to periodically re-score the current
#     best candidate on the full golden set.

from __future__ import annotations

import random
import time
import uuid
from typing import (
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from deepeval.models.base_model import DeepEvalBaseLLM

from deepeval.errors import DeepEvalError
from deepeval.dataset.golden import ConversationalGolden, Golden
from deepeval.optimizer.utils import Aggregator, mean_of_all
from deepeval.optimizer.types import (
    AcceptedIterationDict,
    ModuleId,
    OptimizationReport,
    PromptConfiguration,
    PromptConfigurationId,
    RunnerStatusCallback,
    RunnerStatusType,
    ScoreTable,
)
from deepeval.optimizer.scorer.base import BaseScorer
from deepeval.optimizer.algorithms.base import BaseAlgorithm
from deepeval.optimizer.utils import build_prompt_config_snapshots
from deepeval.prompt.api import PromptType
from deepeval.prompt.prompt import Prompt
from deepeval.optimizer.rewriter import Rewriter

from deepeval.optimizer.algorithms.configs import (
    MIPROV2_MIN_DELTA,
    MIPROV2_REWRITE_INSTRUCTION_MAX_CHARS,
    SIMBA_DEMO_INPUT_MAX_CHARS,
)
from deepeval.optimizer.algorithms.simba.types import SIMBAStrategy


class SIMBA(BaseAlgorithm):
    """
    SIMBA-style cooperative prompt optimization loop with sync/async execution.

    This runner is intentionally low level and does not know about metrics,
    models, or async configs. It relies on a preconfigured Scorer and
    Rewriter, which are typically constructed by PromptOptimizer.

    Parameters
    ----------
    iterations : int
        Total number of optimization trials. Default is 5.
    minibatch_size : int
        Number of examples drawn per iteration. Default is 8.
    random_seed : int, optional
        RNG seed for reproducibility. If None, derived from time.time_ns().
    exploration_probability : float
        Epsilon greedy exploration rate. Default is 0.2.
    full_eval_every : int, optional
        Fully evaluate best candidate every N trials. Default is 5.
    population_size : int
        Maximum number of candidates in the pool. Default is 4.
    proposals_per_step : int
        Number of child prompts proposed per iteration. Default is 4.
    max_demos_per_proposal : int
        Maximum demos from minibatch for APPEND_DEMO strategy. Default is 3.
    """

    name = "SIMBA"
    SINGLE_MODULE_ID: ModuleId = "__module__"

    def __init__(
        self,
        iterations: int = 5,
        minibatch_size: int = 8,
        random_seed: Optional[int] = None,
        exploration_probability: float = 0.2,
        full_eval_every: Optional[int] = 5,
        population_size: int = 4,
        proposals_per_step: int = 4,
        max_demos_per_proposal: int = 3,
        aggregate_instances: Aggregator = mean_of_all,
        scorer: Optional[BaseScorer] = None,
    ) -> None:
        # Validate parameters
        if iterations < 1:
            raise ValueError("iterations must be >= 1")
        if minibatch_size < 1:
            raise ValueError("minibatch_size must be >= 1")
        if exploration_probability < 0.0 or exploration_probability > 1.0:
            raise ValueError(
                "exploration_probability must be >= 0.0 and <= 1.0"
            )
        if full_eval_every is not None and full_eval_every < 1:
            raise ValueError("full_eval_every must be >= 1")
        if population_size < 1:
            raise ValueError("population_size must be >= 1")
        if proposals_per_step < 1:
            raise ValueError("proposals_per_step must be >= 1")
        if max_demos_per_proposal < 0:
            raise ValueError("max_demos_per_proposal must be >= 0")

        self.iterations = iterations
        self.minibatch_size = minibatch_size
        self.exploration_probability = exploration_probability
        self.full_eval_every = full_eval_every
        self.population_size = population_size
        self.proposals_per_step = proposals_per_step
        self.max_demos_per_proposal = max_demos_per_proposal
        self.aggregate_instances = aggregate_instances
        self.scorer = scorer

        if max_demos_per_proposal > 0:
            self._strategies = [
                SIMBAStrategy.APPEND_DEMO,
                SIMBAStrategy.APPEND_RULE,
            ]
        else:
            self._strategies = [SIMBAStrategy.APPEND_RULE]

        # If no seed provided, use time-based seed
        if random_seed is None:
            random_seed = time.time_ns()
        self.random_seed = random_seed
        self.random_state = random.Random(random_seed)

        # Runtime state to be reset between runs
        self.reset_state()

        # Status callback set by PromptOptimizer:
        #   (kind, step_index, total_steps, detail) -> None
        self.status_callback: Optional[RunnerStatusCallback] = None

        # Optimizer model used by the rewriter for prompt mutation.
        # Set by PromptOptimizer.
        self.optimizer_model: Optional["DeepEvalBaseLLM"] = None

        # Lazy-loaded Rewriter set by PromptOptimizer
        self._rewriter: Optional[Rewriter] = None

    ##############
    # Public API #
    ##############

    def execute(
        self,
        prompt: Prompt,
        goldens: Union[List[Golden], List[ConversationalGolden]],
    ) -> Tuple[Prompt, OptimizationReport]:
        """
        Synchronous SIMBA run from a full list of goldens.

        The full goldens set is used both for mini-batched scoring during
        optimization and for a final full evaluation of the best candidate.
        """
        total_goldens = len(goldens)
        if total_goldens < 1:
            raise DeepEvalError(
                "SIMBA prompt optimization requires at least 1 golden, but "
                f"received {total_goldens}. Provide at least one golden to run "
                "the optimizer."
            )

        self._ensure_scorer()
        self.reset_state()

        # Seed candidate pool with the root prompt configuration.
        seed_prompts_by_module = {self.SINGLE_MODULE_ID: prompt}
        root_prompt_configuration = PromptConfiguration.new(
            prompts=dict(seed_prompts_by_module)
        )
        # Add root candidate to the pool, but defer its first minibatch
        # evaluation until the first iteration so that any long running
        # model calls happen under the main loop (with progress updates).
        self._add_prompt_configuration(root_prompt_configuration)

        accepted_iterations: List[Dict] = []
        self.trial_index = 0

        def _one_iteration() -> bool:
            nonlocal accepted_iterations

            if not goldens:
                return False

            # Lazily seed with a minibatch score for the root
            # candidate on the first iteration.
            if not self._minibatch_score_counts:
                seed_minibatch = self._draw_minibatch(goldens)
                root_score = self.scorer.score_minibatch(
                    root_prompt_configuration, seed_minibatch
                )
                self._record_minibatch_score(
                    root_prompt_configuration.id, root_score
                )

            # 1. Choose which candidate prompt to mutate.
            parent_prompt_configuration = self._select_candidate()
            selected_module_id: ModuleId = self.SINGLE_MODULE_ID

            minibatch = self._draw_minibatch(goldens)

            # Compute shared feedback for this parent/minibatch that will be
            # used by all SIMBA proposals in this iteration.
            feedback_text = self.scorer.get_minibatch_feedback(
                parent_prompt_configuration, selected_module_id, minibatch
            )

            before_mean = self._mean_minibatch_score(
                parent_prompt_configuration.id
            )
            jitter = 1e-6
            min_delta = max(MIPROV2_MIN_DELTA, jitter)

            # 2. Generate multiple SIMBA child prompts and evaluate them.
            num_proposals = int(self.proposals_per_step)
            for _ in range(num_proposals):
                strategy = self._sample_strategy()
                child_prompt = self._generate_child_prompt(
                    strategy,
                    selected_module_id,
                    parent_prompt_configuration,
                    feedback_text,
                    minibatch,
                )
                if child_prompt is None:
                    # No child, nothing to evaluate for this proposal.
                    continue

                child_prompt_configuration = self._make_child(
                    selected_module_id,
                    parent_prompt_configuration,
                    child_prompt,
                )

                child_score = self.scorer.score_minibatch(
                    child_prompt_configuration, minibatch
                )

                # 3. Evaluate & decide whether to accept the child.
                if child_score >= before_mean + min_delta:
                    # Accept: add to pool, update surrogate stats, and record iteration.
                    self._add_prompt_configuration(child_prompt_configuration)
                    self._record_minibatch_score(
                        child_prompt_configuration.id, child_score
                    )

                    accepted_iterations.append(
                        AcceptedIterationDict(
                            parent=parent_prompt_configuration.id,
                            child=child_prompt_configuration.id,
                            module=selected_module_id,
                            before=before_mean,
                            after=child_score,
                        )
                    )
                # else: reject; do not add child to the candidate pool.

            self.trial_index += 1
            if (
                self.full_eval_every is not None
                and self.trial_index % self.full_eval_every == 0
            ):
                self._full_evaluate_best(goldens)

            return True

        self._run_loop_iteration(_one_iteration)

        # Ensure at least one candidate has been fully evaluated.
        if not self.pareto_score_table:
            self._full_evaluate_best(goldens)

        best = self._best_by_aggregate()
        prompt_config_snapshots = build_prompt_config_snapshots(
            self.prompt_configurations_by_id
        )
        report = OptimizationReport(
            optimization_id=self.optimization_id,
            best_id=best.id,
            accepted_iterations=accepted_iterations,
            pareto_scores=self.pareto_score_table,
            parents=self.parents_by_id,
            prompt_configurations=prompt_config_snapshots,
        )
        return best.prompts[self.SINGLE_MODULE_ID], report

    async def a_execute(
        self,
        prompt: Prompt,
        goldens: Union[List[Golden], List[ConversationalGolden]],
    ) -> Tuple[Prompt, OptimizationReport]:
        """
        Asynchronous twin of execute().
        """
        total_goldens = len(goldens)
        if total_goldens < 1:
            raise DeepEvalError(
                "SIMBA prompt optimization requires at least 1 golden, but "
                f"received {total_goldens}. Provide at least one golden to run "
                "the optimizer."
            )

        self._ensure_scorer()
        self.reset_state()

        seed_prompts_by_module = {self.SINGLE_MODULE_ID: prompt}
        root_prompt_configuration = PromptConfiguration.new(
            prompts=dict(seed_prompts_by_module)
        )
        self._add_prompt_configuration(root_prompt_configuration)

        accepted_iterations: List[Dict] = []
        self.trial_index = 0

        async def _one_iteration() -> bool:
            nonlocal accepted_iterations

            if not goldens:
                return False

            if not self._minibatch_score_counts:
                seed_minibatch = self._draw_minibatch(goldens)
                root_score = await self.scorer.a_score_minibatch(
                    root_prompt_configuration, seed_minibatch
                )
                self._record_minibatch_score(
                    root_prompt_configuration.id, root_score
                )

            parent_prompt_configuration = self._select_candidate()
            selected_module_id: ModuleId = self.SINGLE_MODULE_ID

            minibatch = self._draw_minibatch(goldens)

            feedback_text = await self.scorer.a_get_minibatch_feedback(
                parent_prompt_configuration, selected_module_id, minibatch
            )

            before_mean = self._mean_minibatch_score(
                parent_prompt_configuration.id
            )
            jitter = 1e-6
            min_delta = max(MIPROV2_MIN_DELTA, jitter)

            num_proposals = int(self.proposals_per_step)
            for _ in range(num_proposals):
                strategy = self._sample_strategy()
                child_prompt = await self._a_generate_child_prompt(
                    strategy,
                    selected_module_id,
                    parent_prompt_configuration,
                    feedback_text,
                    minibatch,
                )
                if child_prompt is None:
                    continue

                child_prompt_configuration = self._make_child(
                    selected_module_id,
                    parent_prompt_configuration,
                    child_prompt,
                )

                child_score = await self.scorer.a_score_minibatch(
                    child_prompt_configuration, minibatch
                )

                if child_score >= before_mean + min_delta:
                    self._add_prompt_configuration(child_prompt_configuration)
                    self._record_minibatch_score(
                        child_prompt_configuration.id, child_score
                    )

                    accepted_iterations.append(
                        AcceptedIterationDict(
                            parent=parent_prompt_configuration.id,
                            child=child_prompt_configuration.id,
                            module=selected_module_id,
                            before=before_mean,
                            after=child_score,
                        )
                    )

            self.trial_index += 1
            if (
                self.full_eval_every is not None
                and self.trial_index % self.full_eval_every == 0
            ):
                await self._a_full_evaluate_best(goldens)

            return True

        await self._a_run_loop_iteration(_one_iteration)

        if not self.pareto_score_table:
            await self._a_full_evaluate_best(goldens)

        best = self._best_by_aggregate()
        prompt_config_snapshots = build_prompt_config_snapshots(
            self.prompt_configurations_by_id
        )
        report = OptimizationReport(
            optimization_id=self.optimization_id,
            best_id=best.id,
            accepted_iterations=accepted_iterations,
            pareto_scores=self.pareto_score_table,
            parents=self.parents_by_id,
            prompt_configurations=prompt_config_snapshots,
        )
        return best.prompts[self.SINGLE_MODULE_ID], report

    ###################
    # State & helpers #
    ###################

    def reset_state(self) -> None:
        self.optimization_id = str(uuid.uuid4())
        self.prompt_configurations_by_id: Dict[
            PromptConfigurationId, PromptConfiguration
        ] = {}
        self.parents_by_id: Dict[
            PromptConfigurationId, Optional[PromptConfigurationId]
        ] = {}
        # For SIMBA we reuse the same field name as GEPA for full-eval scores.
        self.pareto_score_table: ScoreTable = {}

        # Surrogate stats: running mean minibatch scores per candidate.
        self._minibatch_score_sums: Dict[PromptConfigurationId, float] = {}
        self._minibatch_score_counts: Dict[PromptConfigurationId, int] = {}

        # Trial counter (used for full_eval_every).
        self.trial_index: int = 0

    def _ensure_scorer(self) -> None:
        if self.scorer is None:
            raise DeepEvalError(
                "SIMBARunner requires a `scorer`. "
                "Construct one (for example, Scorer) in "
                "PromptOptimizer and assign it to `runner.scorer`."
            )

    def _prompts_equivalent(
        self,
        old_prompt: Prompt,
        new_prompt: Prompt,
    ) -> bool:
        """
        Compare two Prompts for optimization purposes.

        We treat a child as "no change" if:
        - The types differ, or
        - For TEXT: trimmed text_template matches.
        - For LIST: messages_template length, roles, and trimmed content match.
        """

        if new_prompt.type == PromptType.LIST:
            old_msgs = old_prompt.messages_template
            new_msgs = new_prompt.messages_template
            if len(old_msgs) != len(new_msgs):
                return False

            for old_msg, new_msg in zip(old_msgs, new_msgs):
                if old_msg.role != new_msg.role:
                    return False
                if (old_msg.content or "").strip() != (
                    new_msg.content or ""
                ).strip():
                    return False

            return True

        old_txt = (old_prompt.text_template or "").strip()
        new_txt = (new_prompt.text_template or "").strip()
        return new_txt == old_txt

    def _add_prompt_configuration(
        self,
        prompt_configuration: PromptConfiguration,
    ) -> None:
        """
        Add a candidate to the active pool and, if a population limit is set,
        prune the worst-scoring candidates to enforce it.
        """
        self.prompt_configurations_by_id[prompt_configuration.id] = (
            prompt_configuration
        )
        self.parents_by_id[prompt_configuration.id] = (
            prompt_configuration.parent
        )

        # If we exceed the population size, iteratively prune the worst
        # (by mean minibatch score), never removing the current best.
        while len(self.prompt_configurations_by_id) > self.population_size:
            best_id: Optional[PromptConfigurationId] = None
            best_score = float("-inf")
            for cand_id in self.prompt_configurations_by_id.keys():
                mean_score = self._mean_minibatch_score(cand_id)
                if mean_score > best_score:
                    best_score = mean_score
                    best_id = cand_id

            worst_id: Optional[PromptConfigurationId] = None
            worst_score = float("inf")
            for cand_id in self.prompt_configurations_by_id.keys():
                if cand_id == best_id:
                    continue
                mean_score = self._mean_minibatch_score(cand_id)
                if mean_score < worst_score:
                    worst_score = mean_score
                    worst_id = cand_id

            if worst_id is None or worst_id == best_id:
                break

            # Prune the chosen worst candidate from all bookkeeping tables.
            self.prompt_configurations_by_id.pop(worst_id, None)
            self.parents_by_id.pop(worst_id, None)
            self._minibatch_score_sums.pop(worst_id, None)
            self._minibatch_score_counts.pop(worst_id, None)
            self.pareto_score_table.pop(worst_id, None)

    def _record_minibatch_score(
        self,
        prompt_configuration_id: PromptConfigurationId,
        score: float,
    ) -> None:
        self._minibatch_score_sums[prompt_configuration_id] = (
            self._minibatch_score_sums.get(prompt_configuration_id, 0.0)
            + float(score)
        )
        self._minibatch_score_counts[prompt_configuration_id] = (
            self._minibatch_score_counts.get(prompt_configuration_id, 0) + 1
        )

    def _mean_minibatch_score(
        self,
        prompt_configuration_id: PromptConfigurationId,
    ) -> float:
        total = self._minibatch_score_sums.get(prompt_configuration_id, 0.0)
        count = self._minibatch_score_counts.get(prompt_configuration_id, 0)
        if count <= 0:
            # Use a sentinel that will not dominate selection if a scored
            # candidate exists. Root is seeded explicitly in the first iteration.
            return float("-inf")
        return total / count

    def _best_by_minibatch(self) -> PromptConfiguration:
        """
        Return the candidate with the highest mean minibatch score.
        """
        if not self.prompt_configurations_by_id:
            raise DeepEvalError(
                "SIMBARunner has no prompt configurations; this should not happen."
            )

        best_id: Optional[PromptConfigurationId] = None
        best_score = float("-inf")

        for cand_id in self.prompt_configurations_by_id.keys():
            mean_score = self._mean_minibatch_score(cand_id)
            if mean_score > best_score:
                best_score = mean_score
                best_id = cand_id

        if best_id is None:
            # Fallback to the first candidate if all means are -inf.
            best_id = next(iter(self.prompt_configurations_by_id.keys()))

        return self.prompt_configurations_by_id[best_id]

    def _best_by_aggregate(self) -> PromptConfiguration:
        """
        Return the best candidate based on full-eval scores.

        If no full evaluation scores are available (should be rare, but possible if
        full_eval_every is very large and the loop exits early), fall back to
        best-by-minibatch.
        """
        if not self.pareto_score_table:
            return self._best_by_minibatch()

        totals = {
            prompt_configuration_id: self.aggregate_instances(vector)
            for prompt_configuration_id, vector in self.pareto_score_table.items()
        }

        best_ids: List[PromptConfigurationId] = []
        best_val = float("-inf")

        for cand_id, aggregate in totals.items():
            if aggregate > best_val + 1e-12:
                best_val = aggregate
                best_ids = [cand_id]
            elif abs(aggregate - best_val) <= 1e-12:
                best_ids.append(cand_id)

        chosen_id = self.random_state.choice(best_ids)
        return self.prompt_configurations_by_id[chosen_id]

    def _select_candidate(self) -> PromptConfiguration:
        """
        Epsilon-greedy candidate selection:

        - With probability ``exploration_probability``, pick a random candidate.
        - Otherwise, pick the candidate with the highest mean minibatch score.
        """
        if not self.prompt_configurations_by_id:
            raise DeepEvalError(
                "SIMBARunner has no prompt configurations to select from."
            )

        candidate_ids = list(self.prompt_configurations_by_id.keys())
        if not candidate_ids:
            raise DeepEvalError(
                "SIMBARunner has an empty candidate pool; this should not happen."
            )

        eps = float(self.exploration_probability)
        if eps > 0.0 and self.random_state.random() < eps:
            chosen_id = self.random_state.choice(candidate_ids)
        else:
            chosen_id = self._best_by_minibatch().id

        return self.prompt_configurations_by_id[chosen_id]

    def _draw_minibatch(
        self,
        goldens: Union[List[Golden], List[ConversationalGolden]],
    ) -> Union[List[Golden], List[ConversationalGolden]]:
        """
        Determine effective minibatch size, bounded by the available goldens,
        and sample with replacement.
        """
        n = len(goldens)
        if n <= 0:
            return []

        size = min(self.minibatch_size, n)

        return [goldens[self.random_state.randrange(0, n)] for _ in range(size)]

    async def _a_full_evaluate_best(
        self,
        goldens: Union[List[Golden], List[ConversationalGolden]],
    ) -> None:
        if not self.prompt_configurations_by_id:
            return

        best = self._best_by_minibatch()
        if best.id in self.pareto_score_table:
            return

        scores = await self.scorer.a_score_pareto(best, goldens)
        self.pareto_score_table[best.id] = scores

    def _full_evaluate_best(
        self,
        goldens: Union[List[Golden], List[ConversationalGolden]],
    ) -> None:
        if not self.prompt_configurations_by_id:
            return

        best = self._best_by_minibatch()
        if best.id in self.pareto_score_table:
            return

        scores = self.scorer.score_pareto(best, goldens)
        self.pareto_score_table[best.id] = scores

    async def _a_generate_child_prompt(
        self,
        strategy: SIMBAStrategy,
        selected_module_id: ModuleId,
        parent_prompt_configuration: PromptConfiguration,
        feedback_text: str,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> Optional[Prompt]:
        try:
            old_prompt = parent_prompt_configuration.prompts[selected_module_id]
        except KeyError as exc:
            raise DeepEvalError(
                "SIMBARunner expected a prompt for module_id "
                f"{selected_module_id!r} but none was found in the "
                "current prompt configuration."
            ) from exc

        strategy_feedback = self._build_feedback_for_strategy(
            strategy, feedback_text, minibatch
        )

        new_prompt = await self._rewriter.a_rewrite(
            module_id=selected_module_id,
            old_prompt=old_prompt,
            feedback_text=strategy_feedback,
        )

        if old_prompt.type != new_prompt.type or self._prompts_equivalent(
            old_prompt, new_prompt
        ):
            # Don't accept if new prompt is the same as parent, or if type changed.
            return None
        return new_prompt

    def _generate_child_prompt(
        self,
        strategy: SIMBAStrategy,
        selected_module_id: ModuleId,
        parent_prompt_configuration: PromptConfiguration,
        feedback_text: str,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> Optional[Prompt]:
        try:
            old_prompt = parent_prompt_configuration.prompts[selected_module_id]
        except KeyError as exc:
            # This should never happen in normal operation.
            raise DeepEvalError(
                "SIMBARunner expected a prompt for module_id "
                f"{selected_module_id!r} but none was found in the "
                "current prompt configuration."
            ) from exc

        strategy_feedback = self._build_feedback_for_strategy(
            strategy, feedback_text, minibatch
        )

        new_prompt = self._rewriter.rewrite(
            module_id=selected_module_id,
            old_prompt=old_prompt,
            feedback_text=strategy_feedback,
        )

        if old_prompt.type != new_prompt.type or self._prompts_equivalent(
            old_prompt, new_prompt
        ):
            # Don't accept if new prompt is the same as parent, or if type changed.
            return None
        return new_prompt

    def _make_child(
        self,
        selected_module_id: ModuleId,
        parent_prompt_configuration: PromptConfiguration,
        child_prompt: Prompt,
    ) -> PromptConfiguration:
        child_prompt_configuration = PromptConfiguration.new(
            prompts=dict(parent_prompt_configuration.prompts),
            parent=parent_prompt_configuration.id,
        )
        child_prompt_configuration.prompts[selected_module_id] = child_prompt
        return child_prompt_configuration

    def _truncate_instruction(self, text: str) -> str:
        """
        Truncate strategy instructions + feedback to the configured character
        budget so the rewriter prompt does not explode.
        """
        max_chars = MIPROV2_REWRITE_INSTRUCTION_MAX_CHARS
        if max_chars <= 0:
            return text
        if len(text) <= max_chars:
            return text
        return text[:max_chars]

    def _build_demo_block(
        self,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> str:
        """
        Build a small block of input/context/output demos from the current
        minibatch, inspired by SIMBA's `append_a_demo` strategy.

        For each Golden:

            Golden:
                Input   <- golden.input
                Context <- " ".join(golden.context) if present
                Output  <- golden.expected_output

            ConversationalGolden:
                Input   <- golden.scenario
                Context <- " ".join(golden.context) if present
                Output  <- golden.expected_outcome

        All text segments are independently truncated to `SIMBA_DEMO_INPUT_MAX_CHARS`.
        """
        max_demos = self.max_demos_per_proposal
        if max_demos <= 0:
            return ""

        lines: List[str] = []
        demo_limit = min(max_demos, len(minibatch))
        max_chars = SIMBA_DEMO_INPUT_MAX_CHARS

        for golden in minibatch[:demo_limit]:
            if isinstance(golden, Golden):
                input_text = golden.input or ""
                expected_output_text = golden.expected_output or ""
                ctx_list = golden.context or []
            elif isinstance(golden, ConversationalGolden):
                input_text = golden.scenario or ""
                expected_output_text = golden.expected_outcome or ""
                ctx_list = golden.context or []
            else:
                # Unknown type; skip defensively
                continue

            context_text = " ".join(ctx_list) if ctx_list else ""

            # Skip completely empty triples
            if not input_text and not expected_output_text and not context_text:
                continue

            # Truncate each segment independently
            if max_chars > 0:
                if len(input_text) > max_chars:
                    input_text = input_text[:max_chars]
                if len(context_text) > max_chars:
                    context_text = context_text[:max_chars]
                if len(expected_output_text) > max_chars:
                    expected_output_text = expected_output_text[:max_chars]

            demo_lines: List[str] = [f"Input: {input_text}"]
            if context_text:
                demo_lines.append(f"Context: {context_text}")
            demo_lines.append(f"Output: {expected_output_text}")

            lines.append("\n".join(demo_lines))

        return "\n\n".join(lines)

    def _build_feedback_for_strategy(
        self,
        strategy: SIMBAStrategy,
        feedback_text: str,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> str:
        """
        Construct a strategy-specific feedback string that is passed into
        Rewriter.rewrite / a_rewrite.

        - APPEND_RULE: emphasize extracting a concise rule from metric feedback.
        - APPEND_DEMO: emphasize appending concrete demos built from goldens.
        """
        base = (feedback_text or "").strip()

        if strategy is SIMBAStrategy.APPEND_RULE:
            prefix = (
                "Strategy: Append a concise natural-language rule to the existing "
                "prompt that addresses the issues described below. Preserve all "
                "original instructions and add the new rule(s) in a clearly marked "
                '"Rules" or "Guidelines" section.\n\n'
            )
            text = prefix
            if base:
                text += "Evaluation feedback:\n" + base
            return self._truncate_instruction(text)

        if strategy is SIMBAStrategy.APPEND_DEMO:
            demos = self._build_demo_block(minibatch)
            prefix = (
                "Strategy: Append one or more concrete input/output demonstrations "
                "to the prompt. Each demo should illustrate how to respond "
                "correctly on similar inputs.\n\n"
            )
            text = prefix
            if base:
                text += "Evaluation feedback:\n" + base + "\n\n"
            if demos:
                text += (
                    "Candidate demos built from the current minibatch:\n"
                    + demos
                )
            return self._truncate_instruction(text)

        # just pass through feedback.
        return self._truncate_instruction(base)

    def _sample_strategy(self) -> SIMBAStrategy:
        """
        Sample one of the configured SIMBA edit strategies.

        Defaults to APPEND_RULE if the strategy list is empty for any reason.
        """
        return self.random_state.choice(self._strategies)

    def _update_progress(
        self,
        total_iterations: int,
        iteration: int,
        remaining_iterations: int,
        elapsed: float,
    ) -> None:
        if self.status_callback is not None:
            detail = (
                f"(iterations={total_iterations}) "
                f"• iteration {iteration}/{total_iterations} "
                f"• {elapsed:.2f}s • remaining={remaining_iterations}"
            )
            self.status_callback(
                RunnerStatusType.PROGRESS,
                step_index=iteration,
                total_steps=total_iterations,
                detail=detail,
            )

    def _update_error(
        self,
        total_iterations: int,
        iteration: int,
        exc: Exception,
    ) -> None:
        # Report a user-facing error event.
        if self.status_callback is not None:
            detail = (
                f"(iterations={total_iterations}) "
                f"• error {exc.__class__.__name__}: {exc} "
                f"• halted at iteration {iteration}"
            )
            self.status_callback(
                RunnerStatusType.ERROR,
                step_index=iteration,
                total_steps=total_iterations,
                detail=detail,
            )

    def _run_loop_iteration(
        self,
        simba_iteration: Callable[[], bool],
    ) -> None:
        total_iterations = self.iterations
        remaining_iterations = total_iterations
        iteration = 0
        self._update_progress(
            total_iterations, iteration, remaining_iterations, 0.0
        )
        while remaining_iterations > 0:
            iteration += 1
            start_time = time.perf_counter()
            try:
                ok = simba_iteration()
            except Exception as exc:
                self._update_error(total_iterations, iteration, exc)
                break
            elapsed = time.perf_counter() - start_time
            if not ok:
                break
            remaining_iterations -= 1
            self._update_progress(
                total_iterations, iteration, remaining_iterations, elapsed
            )

    async def _a_run_loop_iteration(
        self,
        a_simba_iteration: Callable[[], Awaitable[bool]],
    ) -> None:
        total_iterations = self.iterations
        remaining_iterations = total_iterations
        iteration = 0
        self._update_progress(
            total_iterations, iteration, remaining_iterations, 0.0
        )
        while remaining_iterations > 0:
            iteration += 1
            start_time = time.perf_counter()
            try:
                ok = await a_simba_iteration()
            except Exception as exc:
                self._update_error(total_iterations, iteration, exc)
                break
            elapsed = time.perf_counter() - start_time
            if not ok:
                break
            remaining_iterations -= 1
            self._update_progress(
                total_iterations, iteration, remaining_iterations, elapsed
            )



from typing import Optional


class JsonCorrectnessTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
    """

    @staticmethod
    def generate_reason(
        actual_output: str, expected_schema: str, is_valid_json: bool
    ):
        return f"""Based on the given generated json, generated by an LLM, and a boolean stating whether it is a valid JSON based on the expected json schema, give a reason why it is OR is not a valid Json.

{JsonCorrectnessTemplate.multimodal_rules}

** 
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The generated Json is <is_valid_json> because <your_reason>."
}}

If the json is not a valid one, your reason MUST compare `Expected Json Schema` and `Generated Json` in your reason. Keep it SHORT and CONCISE while being very FACTUAL and ACTIONABLE.
**

Generated Json:
{actual_output}

Expected Json Schema:
{expected_schema}

Is Valid Json?
{is_valid_json}

JSON:
"""



from typing import List, Optional, Literal
from pydantic import BaseModel, Field


class FaithfulnessVerdict(BaseModel):
    reason: Optional[str] = Field(default=None)
    verdict: Literal["yes", "no", "idk"]


class Verdicts(BaseModel):
    verdicts: List[FaithfulnessVerdict]


class Truths(BaseModel):
    truths: List[str]


class Claims(BaseModel):
    claims: List[str]


class FaithfulnessScoreReason(BaseModel):
    reason: str


class InteractionFaithfulnessScore(BaseModel):
    score: float
    reason: Optional[str]
    claims: List[str]
    truths: List[str]
    verdicts: List[FaithfulnessVerdict]



from typing import List
import textwrap


class AnswerRelevancyTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
    """

    @staticmethod
    def generate_statements(actual_output: str, multimodal: bool = False):
        return f"""Given the text, breakdown and generate a list of statements presented. Ambiguous statements and single words can be considered as statements, but only if outside of a coherent statement.

Example:
Example text: 
Our new laptop model features a high-resolution Retina display for crystal-clear visuals. It also includes a fast-charging battery, giving you up to 12 hours of usage on a single charge. For security, we’ve added fingerprint authentication and an encrypted SSD. Plus, every purchase comes with a one-year warranty and 24/7 customer support.

{AnswerRelevancyTemplate.multimodal_rules if multimodal else ""}

{{
    "statements": [
        "The new laptop model has a high-resolution Retina display.",
        "It includes a fast-charging battery with up to 12 hours of usage.",
        "Security features include fingerprint authentication and an encrypted SSD.",
        "Every purchase comes with a one-year warranty.",
        "24/7 customer support is included."
    ]
}}
===== END OF EXAMPLE ======
        
**
IMPORTANT: Please make sure to only return in valid and parseable JSON format, with the "statements" key mapping to a list of strings. No words or explanation are needed. Ensure all strings are closed appropriately. Repair any invalid JSON before you output it.
**

Text:
{actual_output}

JSON:
"""

    @staticmethod
    def generate_verdicts(
        input: str, statements: str, multimodal: bool = False
    ):
        return f"""For the provided list of statements, determine whether each statement is relevant to address the input.
Generate JSON objects with 'verdict' and 'reason' fields.
The 'verdict' should be 'yes' (relevant), 'no' (irrelevant), or 'idk' (ambiguous/supporting information).
Provide 'reason' ONLY for 'no' or 'idk' verdicts.
The statements are from an AI's actual output.

{AnswerRelevancyTemplate.multimodal_rules if multimodal else ""}

**
IMPORTANT: Please make sure to only return in valid and parseable JSON format, with the 'verdicts' key mapping to a list of JSON objects. Ensure all strings are closed appropriately. Repair any invalid JSON before you output it.

Expected JSON format:
{{
    "verdicts": [
        {{
            "verdict": "yes"
        }},
        {{
            "reason": <explanation_for_irrelevance>,
            "verdict": "no"
        }},
        {{
            "reason": <explanation_for_ambiguity>,
            "verdict": "idk"
        }}
    ]  
}}

Generate ONE verdict per statement - number of 'verdicts' MUST equal number of statements.
'verdict' must be STRICTLY 'yes', 'no', or 'idk':
- 'yes': statement is relevant to addressing the input
- 'no': statement is irrelevant to the input  
- 'idk': statement is ambiguous (not directly relevant but could be supporting information)
Provide 'reason' ONLY for 'no' or 'idk' verdicts.
**          

Input:
{input}

Statements:
{statements}

JSON:
"""

    @staticmethod
    def generate_reason(
        irrelevant_statements: List[str],
        input: str,
        score: float,
        multimodal: bool = False,
    ):
        return f"""Given the answer relevancy score, the list of reasons of irrelevant statements made in the actual output, and the input, provide a CONCISE reason for the score. Explain why it is not higher, but also why it is at its current score.
The irrelevant statements represent things in the actual output that is irrelevant to addressing whatever is asked/talked about in the input.
If there is nothing irrelevant, just say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying).

{AnswerRelevancyTemplate.multimodal_rules if multimodal else ""}

**
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason. Ensure all strings are closed appropriately. Repair any invalid JSON before you output it.

Example:
Example JSON:
{{
    "reason": "The score is <answer_relevancy_score> because <your_reason>."
}}
===== END OF EXAMPLE ======
**


Answer Relevancy Score:
{score}

Reasons why the score can't be higher based on irrelevant statements in the actual output:
{irrelevant_statements}

Input:
{input}

JSON:
"""
